/**
  *  \file client/chart/messagemode.cc
  *  \brief Starchart Mode: WMessageChartMode
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chart/messagemode.h"
#include "gfx/gfxutil.h"
#include "gfx/gfx.h"
#include "ui/event.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "client/chartloc.h"

/** Constructor.
    \param text Text to display. */
WMessageChartMode::WMessageChartMode(string_t text)
    : WSimpleChartMode(ChartMode_Message),
      text(text), state(0)
{ }

/** Destructor. */
WMessageChartMode::~WMessageChartMode()
{ }

void
WMessageChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    // Other overlays below our text
    WSimpleChartMode::drawOverlays(can, loc);

    // "Frame" in COLOR_DARK
    // "Text" in COLOR_GRAYSCALE+15 .. COLOR_GRAYSCALE+7
    GfxContext ctx;
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setTextAlign(1, 0);
    ctx.setColor(COLOR_DARK);

    // "Frame"
    GfxPoint sc = loc.getViewport().getCenter();
    sc.y += 10;
    for (int i = -1; i <= +1; ++i)
        for (int j = -1; j <= +1; ++j)
            if (i != 0 || j != 0)
                outText(ctx, sc.x+i, sc.y+j, text);

    // "Text"
    if (state <= TIMER_HZ/2)
        ctx.setColor(COLOR_GRAYSCALE+15);
    else
        ctx.setColor(COLOR_GRAYSCALE+15 - (state - TIMER_HZ/2) * 10 / (TIMER_HZ/2));
    outText(ctx, sc.x, sc.y, text);
}

bool
WMessageChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Tick) {
        // PCC 1.x uses a 18 Hz timer running until 16, i.e. about a second.
        // Our timer runs faster.
        WSimpleChartMode::handleEvent(event, loc);
        ++state;
        if (state > TIMER_HZ) {
            // expired
            delete this;
        } else {
            // still runs. Decay starts after half a second
            if (state > TIMER_HZ/2)
                loc.markChanged();
        }
        return true;
    } else if (event.type == ev_Key || (event.type == ev_Mouse && event.mouse.b != 0)) {
        // These events cause the mode to disappear, but the event
        // is handled regularily.
        WChartMode* m = getNextMode();
        loc.markChanged();
        delete this;
        return m && m->handleEvent(event, loc);
    } else {
        return false;
    }
}

/** Constructor. Creates a WMessageChartMode and adds it to the mode list.
    \param text Text to display
    \param chart_modes Display text on this chart mode set */
void
WMessageChartMode::showMessage(WChartModeList& chart_modes, string_t text)
{
    chart_modes.pushMode(*new WMessageChartMode(text));
}
