/**
  *  \file client/widgets/histlist.cc
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include "client/widgets/histlist.h"
#include "game/team.h"
#include "game/spec.h"
#include "game/ship.h"
#include "game/univ.h"
#include "game/objlist.h"
#include "game/global.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"
#include "gfx/gfxcontext.h"
#include "ui/skin.h"
#include "game/format.h"
#include "ui/util.h"

namespace {
    class AgeSort : public GObjectCompare {
     public:
        AgeSort(const GUniverse& univ, const WHistoryShipSelection& sel)
            : univ(univ), sel(sel)
            { }
        int compare(GObject& a, GObject& b)
            {
                GShip* sa = dynamic_cast<GShip*>(&a);
                GShip* sb = dynamic_cast<GShip*>(&b);
                int ageA = sa != 0 ? sel.getShipLastTurn(univ, sa->getId()) : 0;
                int ageB = sb != 0 ? sel.getShipLastTurn(univ, sb->getId()) : 0;
                return ageA < ageB ? +1 : ageA > ageB ? -1 : 0;
            }
     private:
        const GUniverse& univ;
        const WHistoryShipSelection& sel;
    };

    class AgeDivi : public GObjectDivider {
     public:
        AgeDivi(const GUniverse& univ, const WHistoryShipSelection& sel)
            : univ(univ), sel(sel)
            { }
        string_t getDivider(GObject& a)
            {
                if (GShip* sa = dynamic_cast<GShip*>(&a)) {
                    return format(_("Turn %d"), sel.getShipLastTurn(univ, sa->getId()));
                } else {
                    return _("unknown");
                }
            }
     private:
        const GUniverse& univ;
        const WHistoryShipSelection& sel;
    };
}

/** Build list of history ships. */
void
WHistoryShipSelection::buildList(GObjectList& list, const GUniverse& univ) const
{
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        const GShip& sh = univ.getShip(sid);
        bool accept = false;
        if (sh.getShipKind() != GShip::NoShip) {
            switch (mode) {
             case AllShips:
                accept = true;
                break;
             case LocalShips:
                for (int t = sh.getHistoryNewestTurn(); t > sh.getHistoryOldestTurn(); --t) {
                    const TDbShipTrackEntry e = sh.getHistoryEntry(t);
                    if (e.x > 0 && e.y > 0 && std::abs(e.x - location.x) < 10 && std::abs(e.y - location.y) < 10) {
                        accept = true;
                        break;
                    }
                }
                break;
             case ForeignShips:
                accept = (sh.getOwner() != getPlayerId());
                break;
             case TeamShips:
                accept = (getPlayerTeam(sh.getOwner()) == getPlayerTeam(getPlayerId()));
                break;
             case EnemyShips:
                accept = (getPlayerTeam(sh.getOwner()) != getPlayerTeam(getPlayerId()));
                break;
             case OwnShips:
                accept = (sh.getOwner() == getPlayerId());
                break;
            }
        }
        if (accept) {
            list.addObject(current_histship_type, sid);
        }
    }

    switch (sortOrder) {
     case ById:
        break;

     case ByOwner:
        list.sort(sortByOwner);
        list.addDividers(diviOwner);
        break;

     case ByHull:
        list.sort(sortByHull);
        list.addDividers(diviHull);
        break;

     case ByAge:
        {
            AgeSort s(univ, *this);
            AgeDivi d(univ, *this);
            list.sort(s);
            list.addDividers(d);
        }
        break;

     case ByName:
        list.sort(sortByName);
        break;
    }
}

/** Get number of last turn a ship was seen.
    \param univ Universe to query
    \param sid Ship Id. Must be among the ships returned by buildList.
    \return turn number, 0 if none */
int
WHistoryShipSelection::getShipLastTurn(const GUniverse& univ, int sid) const
{
    const GShip& sh = univ.getShip(sid);
    for (int t = sh.getHistoryNewestTurn(); t > sh.getHistoryOldestTurn(); --t) {
        const TDbShipTrackEntry e = sh.getHistoryEntry(t);
        if (e.x > 0 && e.y > 0
            && (mode != LocalShips
                || (std::abs(e.x - location.x) < 10 && std::abs(e.y - location.y) < 10)))
        {
            return t;
        }
    }
    return 0;
}

/** Get set of visible modes. */
SmallSet<WHistoryShipSelection::Mode>
WHistoryShipSelection::getVisibleModes(const GUniverse& univ) const
{
    SmallSet<Mode> modes;
    SmallSet<Mode> expect = SmallSet<Mode>::allUpTo(Mode(ModeMax));
    bool expectLocalShips = (location.x != 0 || location.y != 0);
    if (!expectLocalShips) {
        expect -= LocalShips;
    }

    /* Check all ships */
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        const GShip& sh = univ.getShip(sid);
        if (sh.getShipKind() != GShip::NoShip) {
            /* We have a ship. Process owner modes */
            modes += AllShips;
            if (sh.getOwner() == getPlayerId()) {
                modes += OwnShips;
            } else {
                modes += ForeignShips;
            }
            if (getPlayerTeam(sh.getOwner()) == getPlayerTeam(getPlayerId())) {
                modes += TeamShips;
            } else {
                modes += EnemyShips;
            }

            /* Check for local mode */
            if (expectLocalShips) {
                for (int t = sh.getHistoryNewestTurn(); t > sh.getHistoryOldestTurn(); --t) {
                    const TDbShipTrackEntry e = sh.getHistoryEntry(t);
                    if (e.x > 0 && e.y > 0 && std::abs(e.x - location.x) < 10 && std::abs(e.y - location.y) < 10) {
                        modes += LocalShips;
                        break;
                    }
                }
            }

            /* Check for early exit when full set found */
            if (modes == expect) {
                break;
            }
        }
    }

    /* Discount team modes when we don't have teams */
    if (!isTeamConfigured()) {
        modes -= TeamShips;
        modes -= EnemyShips;
    }
    return modes;
}

WHistoryShipSelection::Mode
WHistoryShipSelection::getInitialMode(const GUniverse& univ)
{
    SmallSet<Mode> modes = getVisibleModes(univ);
    if (modes.contains(LocalShips)) {
        return LocalShips;
    } else {
        return AllShips;
    }
}


string_t
WHistoryShipSelection::getModeName(Mode mode) const
{
    switch (mode) {
     case AllShips:
        return _("All ships");
     case LocalShips:
        return format(_("Ships near (%d,%d)"), location.x, location.y);
     case ForeignShips:
        return _("Foreign ships");
     case TeamShips:
        return _("Team ships");
     case EnemyShips:
        return _("Enemy ships");
     case OwnShips:
        return _("Own ships");
    }
    return string_t();
}

string_t
WHistoryShipSelection::getModeName() const
{
    return getModeName(mode);
}

string_t
WHistoryShipSelection::getSortName(Sort sort)
{
    switch (sort) {
     case ById:
        return _("Sort by Id");
     case ByOwner:
        return _("Sort by Owner");
     case ByHull:
        return _("Sort by Hull");
     case ByAge:
        return _("Sort by Age of scan");
     case ByName:
        return _("Sort by Name");
    }
    return string_t();
}

string_t
WHistoryShipSelection::getSortName() const
{
    return getSortName(sortOrder);
}

/**************************** WHistoryShipList ***************************/

/** Constructor.
    \param id     Widget Id
    \param lines  Lines to display
    \param width  Width of widget
    \param list   (Storage for) ship list
    \param config Initial configuration */
WHistoryShipList::WHistoryShipList(int id, int lines, int width, GObjectList& list, WHistoryShipSelection& config)
    : WObjectList(id, lines, width, list),
      config(config),
      list(list)
{
    setInfoSpace(fonts[FONT_NORMAL]->getTextWidth(format(_("%d turns ago"), 100)));
}

WHistoryShipList::~WHistoryShipList()
{ }

string_t
WHistoryShipList::getInfoForObject(GObjectReference id)
{
    // Additional information, used by WObjectList::drawPart
    GShip* sh;
    if (id.isValid() && (sh = dynamic_cast<GShip*>(&*id)) != 0) {
        int turn = config.getShipLastTurn(getDisplayedTurn().getCurrentUniverse(), sh->getId());
        if (turn > 0) {
            return formatTurnNumber(turn);
        }
    }
    return string_t();
}

/** Update list. Call this after a change to the configuration to
    update the display. This is intended to be hooked into
    WHistoryShipControl::sig_changed. */
void
WHistoryShipList::updateList()
{
    /* Remember current object */
    GObjectReference current = getCurrentObject();

    /* Build new list */
    list.clear();
    config.buildList(list, getDisplayedTurn().getCurrentUniverse());

    /* Redraw */
    handleListChange(current);
}


/************************** WHistoryShipControl **************************/

WHistoryShipControl::WHistoryShipControl(WHistoryShipSelection& config)
    : UIBaseComplexWidget(),
      btn_d(0, "D", 'd'),
      config(config)
{
    btn_d.sig_fired.add(this, &WHistoryShipControl::onClick);
    addChildWidget(btn_d, getLastChild());
}

WHistoryShipControl::~WHistoryShipControl()
{ }

void
WHistoryShipControl::onClick()
{
    // Show modes and sort orders
    StringList list;

    // - Modes
    SmallSet<WHistoryShipSelection::Mode> modes = config.getVisibleModes(getDisplayedTurn().getCurrentUniverse());
    for (int i = 0; i <= WHistoryShipSelection::ModeMax; ++i) {
        WHistoryShipSelection::Mode m = WHistoryShipSelection::Mode(i);
        if (modes.contains(m)) {
            list.add(1000+i, config.getModeName(m));
        }
    }

    // - Sorts
    for (int i = 0; i <= WHistoryShipSelection::SortMax; ++i) {
        WHistoryShipSelection::Sort s = WHistoryShipSelection::Sort(i);
        list.add(2000+i, config.getSortName(s));
    }

    // Do it
    UIStandardListbox box(0, list.getCount(), getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
    box.scrollToKey(2000 + config.sortOrder);
    if (box.doStandardMenu(btn_d)) {
        int key = box.getKey();
        if (key >= 2000) {
            config.sortOrder = WHistoryShipSelection::Sort(key - 2000);
        } else if (key >= 1000) {
            config.mode = WHistoryShipSelection::Mode(key - 1000);
        } else {
            // what?
        }
        sig_changed.raise();
        drawWidget();
    }
}

void
WHistoryShipControl::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Static);
    ctx.setSolidBackground();
    ctx.setTextAlign(0, 1);

    GfxRect r = getExtent();
    r.consumeX(25);
    r.consumeY((r.h - font_heights[FONT_NORMAL])/2);
    outTextF(ctx, r, config.getModeName());

    can.update(getExtent());
}

void
WHistoryShipControl::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint sz = font_sizes[FONT_NORMAL].scaledBy(15, 1);
    sz.y = std::max(sz.y, 20);
    sz.x += 25;

    info.min_size = info.pref_size = sz;
    info.flags = li_FlexibleH;
}

void
WHistoryShipControl::onResize(GfxRect /*oldsize*/, bool /*by_layout*/)
{
    GfxRect r = getExtent();
    r.y += (r.h - 20)/2;
    r.w = 20;
    r.h = 20;
    btn_d.setExtent(r, true);
}

bool
WHistoryShipControl::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && (e.key.code == ss_Alt + 'd' || e.key.code == '#' || e.key.code == '\\')) {
        onClick();
        return true;
    } else {
        return UIBaseComplexWidget::handleEvent(e, second_pass);
    }
}
