/**
  *  \file game/load-trn.cc
  *  \brief Loading a TRN file
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/load.h"
#include "cpluslib/format.h"
#include "game/cmds.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/racename.h"
#include "game/ship.h"
#include "game/trn.h"
#include "util/console.h"
#include "util/except.h"
#include "util/nls.h"
#include "util/recode.h"

/** Find first command that does not apply to a specific unit.
    \param f The turn file
    \param i Command index
    \param cmd_class f.getCommandType(i)
    \param cmd_id f.getCommandId(i)
    \return index such that all commands [i,return) belong to the same unit,
    i.e. the returned index is the first command for the next unit (or one-past-end) */
static int
findFirstCommandForNextUnit(const GTurnfile& f, int i, const int cmd_class, const int cmd_id)
{
    do {
        ++i;
    } while (i < f.getNumCommands() && cmd_class == f.getCommandType(i) && cmd_id == f.getCommandId(i));
    return i;
}

/** Load turn file.
    \param trn    Game turn data
    \param s      Stream to load (see GTurnfile constructor)
    \param player Player number */
void
loadTurn(GGameTurn& trn, Stream& s, int player)
{
    console.write(LOG_INFO, format(_("Loading %s turn file..."), player_racenames.getAdjName(player)));

    GTurnfile f(s);
    if (f.getFeatures() & GTurnfile::trnf_Taccom)
        console.write(LOG_INFO, format(_("Turn file contains %d attachment%!1{s%}"), f.getNumFiles()));
    f.sortCommands();

    /* Pass 1: verify commands */
    for (int i = 0; i < f.getNumCommands(); ++i) {
        const int cmd_class = f.getCommandType(i);
        const int cmd_id    = f.getCommandId(i);
        switch (cmd_class) {
         case GTurnfile::UndefinedCommand:
            throw FileFormatException(s, format(_("Turn file contains invalid command code %d"), f.getCommandCode(i)));
         case GTurnfile::ShipCommand:
            if (cmd_id <= 0 && cmd_id > NUM_SHIPS)
                throw FileFormatException(s, format(_("Turn file refers to non-existant ship %d"), cmd_id));
            if (!trn.getCurrentUniverse().getShip(cmd_id).getShipSource().contains(player))
                throw FileFormatException(s, format(_("Turn file refers to ship %d which is not ours"), cmd_id));
            break;
         case GTurnfile::PlanetCommand:
            if (cmd_id <= 0 && cmd_id > NUM_PLANETS)
                throw FileFormatException(s, format(_("Turn file refers to non-existant planet %d"), cmd_id));
            if (!trn.getCurrentUniverse().getPlanet(cmd_id).getPlanetSource().contains(player))
                throw FileFormatException(s, format(_("Turn file refers to planet %d which is not ours"), cmd_id));
            break;
         case GTurnfile::BaseCommand:
            if (cmd_id <= 0 && cmd_id > NUM_PLANETS)
                throw FileFormatException(s, format(_("Turn file refers to non-existant starbase %d"), cmd_id));
            if (!trn.getCurrentUniverse().getPlanet(cmd_id).getBaseSource().contains(player))
                throw FileFormatException(s, format(_("Turn file refers to starbase %d which is not ours"), cmd_id));
            break;
         case GTurnfile::OtherCommand:
            break;
        }
    }

    /* Pass 2: process commands. */
    string_t tim_allies;
    int i = 0;
    while (i < f.getNumCommands()) {
        // Get this command's Id and class
        const int cmd_class = f.getCommandType(i);
        const int cmd_id    = f.getCommandId(i);

        switch (cmd_class) {
         case GTurnfile::ShipCommand:
            {
                /* This used to patch the innards of the GShip object directly (using a
                   member function of GShip), but now works on a local copy of the ship
                   structure, making it independant of GShip's actual representation.
                   For efficiency, we first figure out the number of commands applying
                   to this unit, then process them as a block. */
                const int limit = findFirstCommandForNextUnit(f, i, cmd_class, cmd_id);
                TShip ship;
                trn.getCurrentUniverse().getShip(cmd_id).getShipData(ship);
                while (i < limit) {
                    const int code = f.getCommandCode(i);
                    getPartialStructure(f.getCommandData(i), &ship, ship.parser, f.getCommandCodeRecordIndex(code), f.getCommandLength(i));
                    if (code == tcm_ShipChangeFc && i+1 < limit && f.getCommandCode(i+1) == tcm_ShipChangeFc) {
                        // might be THost alliance command
                        const char* data = f.getCommandData(i);
                        const uint16_t hack = getUint16(data);
                        if (hack == 0x6666 /* ff */ || hack == 0x4646 /* FF */ || hack == 0x6565 /* ee */) {
                            tim_allies.append(data, 3);
                        }
                    }
                    ++i;
                }
                trn.getCurrentUniverse().getShip(cmd_id).addShipData(ship, GPlayerSet(player));
            }
            break;

         case GTurnfile::PlanetCommand:
            {
                /* Same thing as for ships. */
                const int limit = findFirstCommandForNextUnit(f, i, cmd_class, cmd_id);
                TPlanet planet;
                trn.getCurrentUniverse().getPlanet(cmd_id).getPlanetData(planet);
                while (i < limit) {
                    const int code = f.getCommandCode(i);
                    if (code == tcm_PlanetBuildBase) {
                        planet.build_base = 1;
                    } else {
                        getPartialStructure(f.getCommandData(i), &planet, planet.parser, f.getCommandCodeRecordIndex(code), f.getCommandLength(i));
                    }
                    ++i;
                }
                trn.getCurrentUniverse().getPlanet(cmd_id).addPlanetData(planet, GPlayerSet(player));
            }
            break;

         case GTurnfile::BaseCommand:
            {
                /* Same thing as for ships. */
                const int limit = findFirstCommandForNextUnit(f, i, cmd_class, cmd_id);
                TStarbase base;
                trn.getCurrentUniverse().getPlanet(cmd_id).getBaseData(base);
                while (i < limit) {
                    const int code = f.getCommandCode(i);
                    getPartialStructure(f.getCommandData(i), &base, base.parser, f.getCommandCodeRecordIndex(code), f.getCommandLength(i));
                    ++i;
                }
                trn.getCurrentUniverse().getPlanet(cmd_id).addBaseData(base, GPlayerSet(player));
            }
            break;

         case GTurnfile::OtherCommand:
            if (f.getCommandCode(i) == tcm_SendMessage) {
                const char* data = f.getCommandData(i);
                /* format of message is:
                     id     = length
                     data+0 = sender
                     data+2 = receiver
                     data+4 = text */
                const int recv = getUint16(data+2);
                if (recv > 0 && recv <= 12)
                    addMessageFromFile(trn, getMessageFromArray(data+4, cmd_id, false),
                                       player, GPlayerSet(recv));
            } else if (f.getCommandCode(i) == tcm_ChangePassword) {
                trn.getGen(player).setPasswordFromTurn(f.getCommandData(i));
            }
            ++i;
        }
    }
    if (tim_allies.length()) {
        trn.getCommands(player).setCommand(GCommand::phc_TAlliance, 0, convertGameToUtf8(tim_allies));
    }
}
