/**
  *  \file game/objl-sort.cc
  *  \brief Sorting of Object Lists
  *
  *  This module provides the sort-related parts of object lists.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/objlist.h"
#include "game/racename.h"
#include "game/ship.h"
#include "util/nls.h"
#include "game/fleet.h"
#include "game/mission.h"
#include "game/global.h"
#include "cpluslib/format.h"
#include "util/unicodechars.h"
#include "game/fcode.h"
#include "game/host.h"

namespace {
    struct SortAdaptor {
        GObjectCompare& cmp;
        bool operator()(GObjectReference l, GObjectReference r) {
            int res = cmp.compare(*l, *r);
            if (res == 0)
                return l->getId() < r->getId();
            else
                return res < 0;
        }
        SortAdaptor(GObjectCompare& cmp)
            : cmp(cmp)
            { }
    };
}

/** Return iterator to first game object of a list. After removing
    dividers, there may still remain non-game objects (such as the
    "Jettison" entry in the cargo transfer selection). */
static GObjectList::container_t::iterator
skipNonGameObjects(GObjectList::container_t& l)
{
    GObjectList::container_t::iterator it = l.begin();
    while (it != l.end() && !it->isGameObject()) {
        ++it;
    }
    return it;
}

/** Sort using a binary comparison function.
    \param func function returning a strcmp-like result (-1 if first comes
    before second, 0 if equal, 1 if comes after. Objects that compare equal
    will be sorted by Id. */
void
GObjectList::sort(int (*func)(GObject&,GObject&))
{
    class Adaptor : public GObjectCompare {
     public:
        int compare(GObject& a, GObject& b)
            { return func(a, b); }
        Adaptor(int (*func)(GObject&,GObject&))
            : func(func)
            { }
     private:
        int (*func)(GObject&,GObject&);
    };
    Adaptor ad(func);
    sort(ad);
}

/** Sort using an unary weight function.
    \param func function returning a weight. Large weights are sorted
    towards the end, small weights come first. Equal weights are sorted
    by Id. */
void
GObjectList::sort(int32_t (*func)(GObject&))
{
    class Adaptor : public GObjectCompare {
     public:
        int compare(GObject& a, GObject& b)
            {
                int32_t va = func(a);
                int32_t vb = func(b);
                return va < vb ? -1 : va > vb ? +1 : 0;
            }
        Adaptor(int32_t (*func)(GObject&))
            : func(func)
            { }
     private:
        int32_t (*func)(GObject&);
    };
    Adaptor ad(func);
    sort(ad);
}

/** Sort using a comparison predicate. */
void
GObjectList::sort(GObjectCompare& cmp)
{
    removeDividers();
    std::sort(skipNonGameObjects(list), list.end(), SortAdaptor(cmp));
    sig_set_changed.raise(0);
}

/***************************** Sort Functions ****************************/

/** \group objlSorters Object List Sorters */
/*@{*/

/** Function for GObjectList::sort: sort by ascending Id number. */
int32
sortById(GObject& o)
{
    return o.getId();
}

/** Function for GObjectList::sort: sort by owner. */
int32
sortByOwner(GObject& o)
{
    return o.getOwner();
}

/** Function for GObjectList::sort: sort by name. */
int
sortByName(GObject& l, GObject& r)
{
    string_t left  = strUCase(l.getName(GObject::PlainName));
    string_t right = strUCase(r.getName(GObject::PlainName));
    return left.compare(right);
}

/** Function for GObjectList::sort: sort by hull. */
int32
sortByHull(GObject& o)
{
    if (GShip* sh = dynamic_cast<GShip*>(&o))
        return sh->getHullId().getRawValue();
    else
        return 0;
}

/** Function for GObjectList::sort: sort by gross mass. */
int32
sortByMass(GObject& o)
{
    if (GShip* sh = dynamic_cast<GShip*>(&o))
        return sh->getMass().getRawValue();
    else
        return 0;
}

/** Function for GObjectList::sort: sort by hull mass. */
int32
sortByHullMass(GObject& o)
{
    if (GShip* sh = dynamic_cast<GShip*>(&o)) {
        mp16_t n = sh->getHullId();
        if (n.isKnown()) {
            return ::getHull(n).getMass();
        }
    }
    return 0;
}

/** Function for GObjectList::sort: sort by damage. */
int32
sortByDamage(GObject& o)
{
    if (GShip* sh = dynamic_cast<GShip*>(&o)) {
        if (sh->getDamage().isKnown()) {
            return sh->getDamage();
        }
    }
    return 0;
}

/** Function for GObjectList::sort: sort by fleet. */
int32
sortByFleet(GObject& o)
{
    if (GShip* sh = dynamic_cast<GShip*>(&o)) {
        int fid = sh->getFleetNumber();
        if (fid != 0) {
            if (fid == sh->getId()) {
                fid = 2*fid;
            } else {
                fid = 2*fid+1;
            }
        }
        return fid;
    } else {
        return 0;
    }
}

/** Function for GObjectList::sort: sort by tow group. */
int32
sortByTowGroup(GObject& o)
{
    if (GShip* sh = dynamic_cast<GShip*>(&o)) {
        /* This is a tower */
        if (sh->getShipKind() == GShip::CurrentShip && sh->getMission() == GMission::msn_Tow && current_anyship_type.isValidIndex(sh->getTowId())) {
            return 2*sh->getTowId();
        }
        /* Is this ship towed? */
        for (int i = current_ship_type.findNextIndex(0); i != 0; i = current_ship_type.findNextIndex(i)) {
            const GShip* other = dynamic_cast<GShip*>(&current_ship_type.getObjectByIndex(i));
            if (other != 0 && other->getMission() == GMission::msn_Tow && other->getTowId() == sh->getId()) {
                return 2*sh->getId() + 1;
            }
        }
    }
    return 0;
}

int
sortByBattleOrder(GObject& o)
{
    const int UNKNOWN = 1200;
    if (GShip* sh = dynamic_cast<GShip*>(&o)) {
        if (!sh->isFCodeKnown()) {
            return UNKNOWN;
        }

        int val = GFCode::getNumericValue(sh->getFCode());
        if (host.isPHost()) {
            if (val == 1000) {
                // 4.1e rule: Kill gets 1000, capital gets 1002, freighters get 1004.
                // (before: capital gets 1000, freighters get 1002)
                if (sh->getMission() == GMission::msn_Kill) {
                    val = 1000;
                } else if (sh->getNumBeams() > 0 || sh->getNumTorpLaunchers() > 0 || sh->getNumBays() > 0) {
                    val = 1002;
                } else {
                    val = 1004;
                }
            }
        } else {
            if (val < 0) {
                val = 1000;
            }
            if (val == 1000) {
                if (sh->getMission() == GMission::msn_Kill) {
                    val += 10;
                }
                if (sh->getPrimaryEnemy() != 0) {
                    val += 5;
                }
            }
        }
        return val;
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(&o)) {
        if (!pl->isFCodeKnown()) {
            return UNKNOWN;
        }
        if (!host.isPHost()) {
            // Planets have no battle order in non-PHost
            return UNKNOWN;
        }

        int val;
        if (pl->getFCode() == "ATT" || pl->getFCode() == "NUK") {
            val = 0;
        } else {
            val = GFCode::getNumericValue(pl->getFCode());
            if (val == 1000) {
                if (pl->getStructures(str_Defense)) {
                    val = 1001;
                } else {
                    val = 1003;
                }
            }
        }
        return val;
    } else {
        return UNKNOWN;
    }
}

int
sortByLocation(GObject& l, GObject& r)
{
    GMapObject* ml = dynamic_cast<GMapObject*>(&l);
    GMapObject* mr = dynamic_cast<GMapObject*>(&r);

    GPoint pl = ml != 0 ? ml->getPos() : GPoint(0, 0);
    GPoint pr = mr != 0 ? mr->getPos() : GPoint(0, 0);

    return pr.x != pl.x
        ? pr.x - pl.x
        : pr.y - pl.y;
}

int
sortByNewLocation(GObject& l, GObject& r)
{
    GShip* sl = dynamic_cast<GShip*>(&l);
    GShip* sr = dynamic_cast<GShip*>(&r);

    GPoint pl = sl != 0 ? sl->getPredictedPos() : GPoint(0, 0);
    GPoint pr = sr != 0 ? sr->getPredictedPos() : GPoint(0, 0);

    return pr.x != pl.x
        ? pr.x - pl.x
        : pr.y - pl.y;
}

/*@}*/

/*************************** Divider Functions ***************************/

/** \group objlDividers Object List Divider Functions */
/*@{*/

/** Function for GObjectList::addDividers: show owners. Input should
    already be sorted by owner. */
string_t
diviOwner(GObject& o)
{
    return player_racenames.getShortName(o.getOwner());
}

/** Function for GObjectList::addDividers: show hull types. */
string_t
diviHull(GObject& o)
{
    if (GShip* sh = dynamic_cast<GShip*>(&o)) {
        mp16_t n = sh->getHullId();
        if (n.isKnown()) {
            return ::getHull(n).getName();
        }
    }
    return _("unknown");
}

/** Function for GObjectList::addDividers: show fleet names. */
string_t
diviFleet(GObject& o)
{
    int32 key = sortByFleet(o) >> 1;
    if (key != 0) {
        return getFleetTitle(getDisplayedTurn().getCurrentUniverse(), key);
    } else {
        return _("not in a fleet");
    }
}

/** Function for GObjectList::addDividers: show tow group names. */
string_t
diviTowGroup(GObject& o)
{
    int32 key = sortByTowGroup(o) >> 1;
    if (current_anyship_type.isValidIndex(key)) {
        return format(_("towing %s"), current_anyship_type.getObjectByIndex(key).getName(GObject::PlainName));
    } else {
        return _("not in a tow group");
    }
}

string_t
diviBattleOrder(GObject& o)
{
    int n = sortByBattleOrder(o);
    if (n < 0) {
        return "< 0";
    } else if (n < 1000) {
        int level = n/100;
        return format("%d .. %d", 100*level, 100*level+99);
    } else if (n < 1100) {
        return UTF_GEQ " 1000";
    } else {
        return _("unknown");
    }
}

string_t
diviLocation(GObject& o)
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(&o)) {
        return format("(%d,%d)", mo->getPos().x, mo->getPos().y);
    } else {
        return _("not on map");
    }
}

string_t
diviNewLocation(GObject& o)
{
    if (GShip* sh = dynamic_cast<GShip*>(&o)) {
        return format("(%d,%d)", sh->getPredictedPos().x, sh->getPredictedPos().y);
    } else {
        // Just assume it does not move
        return diviLocation(o);
    }
}

/*@}*/
