/**
  *  \file game/parser.h
  *  \brief Message Parser Data Structures
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_PARSER_H
#define PCC_V2_GAME_PARSER_H

#include <vector>
#include <utility>
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "cpluslib/object.h"
#include "cpluslib/ptrvector.h"
#include "game/alliance.h"

/** Object of a message information.
    Together with an Id, determines the target of the information. */
enum GMessageObject {
    MessageShip,                ///< Target is ship given by Id.
    MessagePlanet,              ///< Target is planet given by Id.
    MessageStarbase,            ///< Target is starbase given by Id.
    MessageMinefield,           ///< Target is minefield given by Id.
    MessageIonStorm,            ///< Target is ion storm given by Id.
    MessageExplosion,           ///< This message contains an explosion.
    MessageConfig,              ///< This message contains configuration information.
    MessagePlayerScore,         ///< This message contains player scores. Id is optional and gives the util.dat score Id.
    MessageAlliance,            ///< This message contains alliances.
    MessageNoObject
};

/** Base class of a message value. */
typedef Object GMessageValueBase;

/** Message value.
    A message value contains an index that identifies the value,
    and the actual value.
    \param Index type of index
    \param Value type of value */
template<typename Index, typename Value>
class GMessageValue : public GMessageValueBase {
    const Index index;
    Value value;
 public:
    GMessageValue(Index index, Value value);
    virtual ~GMessageValue();

    Index getIndex() const
        { return index; }

    Value getValue() const
        { return value; }

    void setValue(Value v)
        { this->value = v; }
};

/** Message index for object string attributes. */
enum GMessageStringIndex {
    ms_Name,
    ms_FCode,
    ms_Max
};

/** Message index for object integer attributes. */
enum GMessageIntIndex {
    /* Generic: */
    mi_X,
    mi_Y,
    mi_Radius,
    mi_Owner,
    mi_Damage,
    mi_Heading,
    mi_Speed,

    /* Minefields: */
    mi_MineUnits,
    mi_MineScanReason,
    mi_MineType,
    mi_MineUnitsRemoved,

    /* Ships: */
    mi_ShipHull,
    mi_ShipFuel,
    mi_ShipRemoteFlag,

    /* Planets: */
    mi_PlanetTotalN,            // Total minerals, as in Dark Sense or Super Spy
    mi_PlanetTotalT,
    mi_PlanetTotalD,
    mi_PlanetTotalM,
    mi_PlanetAddedN,            // Added minerals, as in Meteor
    mi_PlanetAddedT,
    mi_PlanetAddedD,
    mi_PlanetAddedM,
    mi_PlanetMinedN,            // Mined minerals, as in Allied Planet
    mi_PlanetMinedT,
    mi_PlanetMinedD,
    mi_PlanetMinedM,
    mi_PlanetCash,
    mi_PlanetSupplies,
    mi_PlanetHasBase,
    mi_PlanetMines,
    mi_PlanetFactories,
    mi_PlanetDefense,
    mi_PlanetTemperature,
    mi_PlanetColonists,
    mi_PlanetActivity,
    mi_PlanetNativeRace,
    mi_PlanetNativeGov,
    mi_PlanetNatives,
    mi_PlanetHasNatives,

    /* Bases: */
    mi_BaseQueuePos,
    mi_BaseQueuePriority,

    /* Score: */
    mi_ScoreWinLimit,
    mi_ScoreTurnLimit,

    /* Storm: */
    mi_IonVoltage,
    mi_IonStatus,

    // TODO: TUtil7Battle
    // TODO: TUtil10Target
    // TODO: TUtil14Wormhole
    // TODO: TUtil17Ion
    // TODO: TUtil33GO
    // TODO: TUtil34FTP
    // TODO: TUtil38PAL
    // TODO: TUtil42GODestroyed
    // TODO: TUtil43MinefieldQuota
    // TODO: TUtil47NonExistantPlanets
    // TODO: Scores
    // TODO: TUtil57Special

    mi_Max
};

/** Object string attribute.
    An arbitrary history information for game objects. */
typedef GMessageValue<GMessageStringIndex,string_t> GMessageStringValue;

/** Object integer attribute.
    An arbitrary history information for game objects. */
typedef GMessageValue<GMessageIntIndex,int32_t> GMessageIntValue;

/** Configuration value. For use with MessageConfig.
    The index is the configuration item name,
    the value is the configuration value. */
typedef GMessageValue<string_t,string_t> GMessageConfigValue;

/** Score value. For use with MessagePlayerScore.
    The index is the player number,
    the value is the score. */
typedef GMessageValue<int,int32_t> GMessageScoreValue;

/** Alliance value. For use with MessageAlliance.
    The string is the alliance identifier.
    The GAllianceOffer is partially-filled in. */
typedef GMessageValue<string_t,GAllianceOffer> GMessageAllianceValue;

/** Message information.
    Collects a set of values for a single object. */
class GMessageInformation {
    typedef ptr_vector<GMessageValueBase> values_t;
    const GMessageObject obj;
    const int32_t  id;
    values_t       values;
    const int      turn;

 public:
    typedef values_t::const_iterator iterator, const_iterator;

    GMessageInformation(GMessageObject obj, int32_t id, int turn);
    ~GMessageInformation();

    void addValue(GMessageStringIndex si, const string_t& s);
    void addValue(GMessageIntIndex ii, int32_t i);
    void addConfigValue(string_t key, string_t value);
    void addScoreValue(int player, int32_t value);
    void addAllianceValue(string_t id, const GAllianceOffer& offer);

    iterator begin() const
        { return values.begin(); }
    iterator end() const
        { return values.end(); }

    GMessageObject getObjectType() const
        { return obj; }
    int32 getObjectId() const
        { return id; }
    int getTurnNumber() const
        { return turn; }
};


template<typename Index, typename Value>
inline GMessageValue<Index,Value>::GMessageValue(Index index, Value value)
    : index(index), value(value)
{ }

template<typename Index, typename Value>
GMessageValue<Index,Value>::~GMessageValue()
{ }

/* game/parser-name.cc: */
const char* getNameForItem(GMessageStringIndex si);
const char* getNameForItem(GMessageIntIndex ii);

GMessageStringIndex getStringIndexFromKeyword(string_t kw);
GMessageIntIndex getIntIndexFromKeyword(string_t kw);

#endif
