/**
  *  \file game/stat.cc
  *  \brief Score Statistics - Management and I/O
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/stat.h"
#include "io/io.h"
#include "io/streams.h"
#include "game/struct.h"
#include "util/nls.h"
#include "util/except.h"
#include "cpluslib/assert.h"
#include "game/gen.h"
#include "game/parser.h"
#include "util/recode.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"

static const char SCORE_FILE_SIG[] = { 'C', 'C', 's', 't', 'a', 't', '0', 26 };

/** Constructor.
    \param turn        Turn number
    \param time        Timestamp
    \param num_fields  Number of fields to allocate */
GStatRecord::GStatRecord(int turn, GTimestamp time, uint16_t num_fields)
    : turn(turn), timestamp(time), scores(num_fields*NUM_PLAYERS, int32_t(-1))
{ }

/** Destructor. */
GStatRecord::~GStatRecord()
{ }

/** Load data from stream. Loads just the 11*num_fields data words. */
void
GStatRecord::loadData(Stream& s)
{
    // Really simple version. Optimize it some day.
    for (GStatIndex i = 0; i < scores.size(); ++i) {
        TInt32 tmp;
        getStructureT(s, tmp);
        scores[i] = tmp.value;
    }
}

/** Save to stream. Saves a header and the data. */
void
GStatRecord::save(Stream& s) const
{
    TScoreRecordHeader header;
    header.turn = turn;
    timestamp.storeRawData(header.timestamp);
    storeStructureT(s, header);
    for (GStatIndex i = 0; i < scores.size(); ++i) {
        TInt32 tmp;
        tmp.value = scores[i];
        storeStructureT(s, tmp);
    }
}

/** Change number of data fields. New fields will be -1 i.e. unknown. */
void
GStatRecord::setNumberOfFields(uint16_t new_count)
{
    scores.resize(new_count*NUM_PLAYERS, int32_t(-1));
}

/******************************* GStatFile *******************************/

const GStatIndex GStatFile::nil;

/** Constructor.
    Makes a blank statistics file with a default schema. */
GStatFile::GStatFile()
    : record_defs(),
      score_descriptions(),
      file_used_future_features(false),
      loaded_records()
{
    init();
}

void
GStatFile::clear()
{
    std::vector<int16>().swap(record_defs);
    std::vector<TScoreDescription>().swap(score_descriptions);
    ptr_vector<GStatRecord>().swap(loaded_records);
    file_used_future_features = false;
}

void
GStatFile::init()
{
    clear();

    /* Build standard schema */
    record_defs.reserve(5);
    record_defs.push_back(ScoreId_Planets);
    record_defs.push_back(ScoreId_Capital);
    record_defs.push_back(ScoreId_Freighters);
    record_defs.push_back(ScoreId_Bases);
    record_defs.push_back(ScoreId_BuildPoints);
}

/** Constructor.
    Loads a statistics file. */
void
GStatFile::load(Stream& s)
{
    clear();

    char tmp[2];
    TScoreHeader new_header;

    /* Try to read file */
    getStructureT(s, new_header);
    if (std::memcmp(new_header.signature, SCORE_FILE_SIG, sizeof(new_header.signature)) != 0)
        throw FileFormatException(s, _("File is missing required signature"));
    if (new_header.header_size < new_header.size
        || new_header.header_fields < 2
        || new_header.record_header < TScoreRecordHeader::size
        || new_header.subfield[0] < new_header.size
        || new_header.subfield[1] < new_header.size)
    {
        throw FileFormatException(s, _("Unsupported file format"));
    }

    if (new_header.header_fields > 2 || new_header.record_header > TScoreRecordHeader::size)
        file_used_future_features = true;

    /* Read record description */
    s.seek(new_header.subfield[0]);
    for (uint32_t i = 0; i < new_header.record_fields; ++i) {
        s.readT(tmp, 2);
        record_defs.push_back(getInt16(tmp));
    }

    /* Read score descriptions */
    s.seek(new_header.subfield[1]);
    s.readT(tmp, 2);
    const uint16_t num_descriptions = getUint16(tmp);
    for (uint32_t i = 0; i < num_descriptions; ++i) {
        TScoreDescription desc;
        getStructureT(s, desc);
        score_descriptions.push_back(desc);
    }

    /* Read scores */
    s.seek(new_header.header_size);
    loaded_records.reserve(new_header.entries);
    for (uint32_t i = 0; i < new_header.entries; ++i) {
        TScoreRecordHeader rh;
        getStructureT(s, rh);
        if (new_header.record_header > TScoreRecordHeader::size)
            s.seek(s.getPos() + (new_header.record_header - TScoreRecordHeader::size));

        GStatRecord* r = new GStatRecord(rh.turn,
                                         rh.timestamp,
                                         new_header.record_fields);
        loaded_records.push_back(r);
        r->loadData(s);
    }
}

/** Load old-style (PCC1.x) statistics file. */
void
GStatFile::loadOldFile(Stream& s)
{
    /* start with default schema */
    init();

    TStatHeader h;
    getStructureT(s, h);
    if (std::memcmp(h.signature, "CC-Stat\032", sizeof(h.signature)) != 0
        || h.entry_size < TStatRecord::size
        || h.entries < 0)
    {
        throw FileFormatException(s, _("Unsupported file format"));
    }

    /* figure out positions (since we use the default schema, we could also use
       constants, but it's more flexible this way) */
    const GStatIndex pi = getAddIndexForScore(ScoreId_Planets);
    const GStatIndex ci = getAddIndexForScore(ScoreId_Capital);
    const GStatIndex fi = getAddIndexForScore(ScoreId_Freighters);
    const GStatIndex bi = getAddIndexForScore(ScoreId_Bases);
    const GStatIndex qi = getAddIndexForScore(ScoreId_BuildPoints);

    /* read individual records */
    for (int32_t i = 0; i < h.entries; ++i) {
        TStatRecord r;
        getStructureT(s, r);
        if (h.entry_size != TStatRecord::size)
            s.seek(s.getPos() + h.entry_size - TStatRecord::size);

        GStatRecord& sr = getRecordForTurn(r.header.turn, r.header.timestamp);
        for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
            sr(pi, pl) = r.scores[pl-1].num_planets;
            sr(ci, pl) = r.scores[pl-1].num_capital;
            sr(fi, pl) = r.scores[pl-1].num_freighters;
            sr(bi, pl) = r.scores[pl-1].num_bases;
            sr(qi, pl) = r.pbps[pl-1];
        }
    }
}

/** Destructor. */
GStatFile::~GStatFile()
{ }

/** Save file. Writes the complete fileto the specified stream.
    This will save the file even if canSaveSafely() returns false; in this
    case, the new copy will contain less information than the file this
    was loaded from. */
void
GStatFile::save(Stream& s)
{
    using std::memcpy;
    char tmp[2];
    long start = s.getPos();

    /* Write preliminary header */
    TScoreHeader header;
    zeroFill(header);
    storeStructureT(s, header);

    /* Write section 1: record definitions */
    header.subfield[0]   = s.getPos();
    header.record_fields = record_defs.size();
    for (uint32_t i = 0; i < record_defs.size(); ++i) {
        storeInt16(tmp, record_defs[i]);
        s.writeT(tmp, 2);
    }

    /* Write section 2: score definitions */
    header.subfield[1] = s.getPos();
    storeInt16(tmp, score_descriptions.size());
    s.writeT(tmp, 2);
    for (uint32_t i = 0; i < score_descriptions.size(); ++i)
        storeStructureT(s, score_descriptions[i]);

    /* Write data */
    header.header_size   = s.getPos();
    header.header_fields = 2;
    header.entries       = loaded_records.size();
    header.record_header = TScoreRecordHeader::size;
    for (uint32_t i = 0; i < loaded_records.size(); ++i)
        loaded_records[i]->save(s);

    /* Write header again */
    memcpy(header.signature, SCORE_FILE_SIG, sizeof(header.signature));
    s.seek(start);
    storeStructureT(s, header);
}

/** Get index for a particular score.
    \param score_id Score identifier, ScoreId_xxx.
    \return Index (>= 0) or nil */
GStatIndex
GStatFile::getIndexForScore(int16_t score_id) const
{
    for (GStatIndex i = 0; i < record_defs.size(); ++i)
        if (record_defs[i] == score_id)
            return i;
    return nil;
}

/** Get index for particular score, or add it. If the score does not exist yet,
    changes the schema to add it.
    \param score_id Score identifier, ScoreId_xxx.
    \return Index (>= 0) */
GStatIndex
GStatFile::getAddIndexForScore(int16_t score_id)
{
    GStatIndex i = getIndexForScore(score_id);
    if (i != nil)
        return i;

    /* we don't have this score, so let's add it. We must change all score records. */
    record_defs.push_back(score_id);
    for (uint32_t n = 0; n < loaded_records.size(); ++n)
        loaded_records[n]->setNumberOfFields(record_defs.size());
    return record_defs.size()-1;
}

/** Get description for a particular score Id.
    \param score_id Score identifier, ScoreId_xxx
    \return pointer to description, or null if none stored in file */
const TScoreDescription*
GStatFile::getDescriptionForScoreId(int16_t score_id) const
{
    for (uint32_t i = 0; i < score_descriptions.size(); ++i)
        if (score_descriptions[i].score_type == score_id)
            return &score_descriptions[i];
    return 0;
}

/** Add score description. If the description already exists, it is overwritten.
    \param desc Description to add
    \return true if this adds or changes a description, false if the new
    description is identical to an existing one */
bool
GStatFile::addScoreDescription(const TScoreDescription& desc)
{
    for (uint32_t i = 0; i < score_descriptions.size(); ++i) {
        TScoreDescription& existing = score_descriptions[i];
        if (existing.score_type == desc.score_type) {
            if (existing.turn_limit == desc.turn_limit
                && existing.win_limit == desc.win_limit
                && std::memcmp(existing.name, desc.name, sizeof(existing.name)))
            {
                return false;
            } else {
                existing = desc;
                return true;
            }
        }
    }
    score_descriptions.push_back(desc);
    return true;
}

/** Get score record for a turn. If none exists, adds one.
    \param turn Turn number (used as key)
    \param ts   Timestamp
    \return reference to record */
GStatRecord&
GStatFile::getRecordForTurn(const int turn, const GTimestamp& ts)
{
    uint32_t index = loaded_records.size();
    while (index > 0 && turn < loaded_records[index-1]->getTurnNumber())
        --index;

    /* Three cases:
       - index = 0 {and turn < [index].turn}
       - turn  = [index-1].turn
       - turn  > [index-1].turn {and turn < [index].turn} */
    if (index > 0 && loaded_records[index-1]->getTurnNumber() == turn) {
        /* If the timestamp differs from the one we know, this is probably
           the result of a re-host. Discard the current scores and start
           anew. */
        if (ts != loaded_records[index-1]->getTimestamp())
            loaded_records.replace_elem(index-1, new GStatRecord(turn, ts, getNumScores()));
        return *loaded_records[index-1];
    }

    /* Only cases 1 and 3 remain. We will make a new record and store that
       at [index]. Shift everything to the right first to make room.
       For simplicity, we append the new record at the end, and shift it
       back; this saves quite some exception hassle. */
    loaded_records.push_back(new GStatRecord(turn, ts, getNumScores()));
    if (loaded_records.size() > 1) {
        for (uint32_t i = loaded_records.size()-1; i > index; --i) {
            /* The last operation this will perform is swapping index-1 and index */
            loaded_records.swap_elems(i-1, i);
        }
    }

    ASSERT(loaded_records[index]->getTurnNumber() == turn);
    return *loaded_records[index];
}

/** Assimilate GEN record into database. */
void
GStatFile::mergeGen(const GGen& g)
{
    GStatRecord& r = getRecordForTurn(g.getTurnNumber(), g.getTimestamp());

    const GStatIndex pi = getAddIndexForScore(ScoreId_Planets);
    const GStatIndex ci = getAddIndexForScore(ScoreId_Capital);
    const GStatIndex fi = getAddIndexForScore(ScoreId_Freighters);
    const GStatIndex bi = getAddIndexForScore(ScoreId_Bases);

    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        // FIXME: implement some sensible merging here.
        // - when we know a score blanker is in use, treat 0 as unknown
        // - do not overwrite a nonzero value with 0 [but what about rehosts?]
        const TGenScore& gs = g.getGenRecord().scores[i-1];
        r(pi, i) = gs.num_planets;
        r(ci, i) = gs.num_capital;
        r(fi, i) = gs.num_freighters;
        r(bi, i) = gs.num_bases;
    }
}

/** Assimilate message information into database. Unlike addScoreDescription, this can
    deal with partial information.
    \param mi Message information
    \param ts Timestamp associated with mi.getTurnNumber() */
void
GStatFile::addMessageInformation(const GMessageInformation& mi, const GTimestamp& ts)
{
    // mi contains:
    // - turn number
    // - optional Id
    // - optional GMessageStringValue: ms_Name
    // - optional GMessageIntValue: mi_ScoreWinLimit, mi_ScoreTurnLimit
    // - optional GMessageScoreValue's

    /* Pass 1: figure out TScoreDescription */
    enum { haveName = 1, haveWinLimit = 2, haveTurnLimit = 4 };
    int have = 0;
    TScoreDescription desc;
    desc.score_type = mi.getObjectId();
    desc.win_limit  = ~0;
    desc.turn_limit = ~0;
    for (GMessageInformation::iterator i = mi.begin(); i != mi.end(); ++i) {
        if (GMessageIntValue* iv = dynamic_cast<GMessageIntValue*>(*i)) {
            if (iv->getIndex() == mi_ScoreWinLimit) {
                desc.win_limit = iv->getValue();
                have |= haveWinLimit;
            }
            if (iv->getIndex() == mi_ScoreTurnLimit) {
                desc.turn_limit = iv->getValue();
                have |= haveTurnLimit;
            }
        } else if (GMessageStringValue* sv = dynamic_cast<GMessageStringValue*>(*i)) {
            if (sv->getIndex() == ms_Name) {
                storeBasicString(desc.name, convertUtf8ToGame(sv->getValue()));
                have |= haveName;
            }
        }
    }

    /* Find existing score description */
    const TScoreDescription* exist = 0;
    for (uint32_t i = 0; i < score_descriptions.size(); ++i) {
        if (desc.score_type != 0 && score_descriptions[i].score_type == desc.score_type) {
            exist = &score_descriptions[i];
            break;
        }
        if ((have & haveName) != 0 && std::memcmp(score_descriptions[i].name, desc.name, sizeof(desc.name)) == 0) {
            exist = &score_descriptions[i];
            break;
        }
    }

    /* If we have no description, make one */
    if (exist == 0) {
        if (desc.score_type != 0) {
            /* We have an identifier. Everything that could be missing by now could be a name. */
            if ((have & haveName) == 0) {
                storeBasicString(desc.name, format(_("Score #%d"), desc.score_type));
            }
        } else if ((have & haveName) != 0) {
            /* We have a name. Allocate an Id. */
            uint16_t i = 1000;
            while (getDescriptionForScoreId(i) != 0)
                ++i;
            desc.score_type = i;
        } else {
            /* Neither name nor Id, we cannot do anything here */
            return;
        }
    } else {
        /* We have a description, so fill in the blanks */
        if ((have & haveName) == 0)
            std::memcpy(desc.name, exist->name, sizeof(desc.name));
        if ((have & haveWinLimit) == 0)
            desc.win_limit = exist->win_limit;
        if ((have & haveTurnLimit) == 0)
            desc.turn_limit = exist->turn_limit;
        desc.score_type = exist->score_type;
    }

    /* Now we can add it. This will overwrite a possibly existing definition with our new,
       updated one. */
    addScoreDescription(desc);

    /* Pass 2: fill in the scores */
    GStatIndex si = getAddIndexForScore(desc.score_type);
    GStatRecord& rec = getRecordForTurn(mi.getTurnNumber(), ts);
    for (GMessageInformation::iterator i = mi.begin(); i != mi.end(); ++i) {
        if (GMessageScoreValue* scv = dynamic_cast<GMessageScoreValue*>(*i)) {
            if (scv->getIndex() > 0 && scv->getIndex() <= NUM_PLAYERS) {
                rec(si, scv->getIndex()) = scv->getValue();
            }
        }
    }
}
