/**
  *  \file int/exec.cc
  *  \brief Interpreter: Bytecode Process Execution
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include <algorithm>
#include "int/exec.h"
#include "int/interror.h"
#include "int/binary.h"
#include "int/unary.h"
#include "int/context.h"
#include "int/compound.h"
#include "int/intglobal.h"
#include "int/csource.h"
#include "int/statement.h"
#include "int/parseexpr.h"
#include "game/gamedirs.h"      // FIXME: game_file_dir
#include "io/streams.h"
#include "io/textfile.h"
#include "util/console.h"
#include "int/ternary.h"
#include "int/args.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "int/optimize.h"
#include "int/array.h"
#include "int/hash.h"
#include "int/notify.h"
#include "int/mutex.h"
#include "int/structure.h"
#include "int/subr.h"

std::auto_ptr<IntContext> IntExecutionContext::global_context;

static bool
noBreak()
{
    return false;
}

static bool (*check_break)() = noBreak;


static void
validateCalledObject(bool isProcedure, uint8_t minor)
{
    if ((minor & IntOpcode::miIMRefuseFunctions) != 0 && !isProcedure)
        throw IntError::typeError(IntError::tyProcedure);
    if ((minor & IntOpcode::miIMRefuseProcedures) != 0 && isProcedure)
        throw IntError::typeError(IntError::tyIndexable);
}

/*************************** IntExecutionFrame ***************************/

/** Construct empty stack frame. */
IntExecutionFrame::IntExecutionFrame()
    : bco(0),
      pc(0),
      local_names(),
      context_sp(0),
      exception_sp(0),
      frame_sp(0),
      want_result(false)
{ }

/** Construct stack frame.
    \param bco Byte code object associated with this frame (=code to execute) */
IntExecutionFrame::IntExecutionFrame(IntBCORef bco)
    : bco(bco),
      pc(0),
      local_names(bco->getLocalNames()),
      context_sp(0),
      exception_sp(0),
      frame_sp(0),
      want_result(false)
{ }

/** Destroy stack frame. */
IntExecutionFrame::~IntExecutionFrame()
{ }

/************************** IntExecutionContext **************************/

/** Constructor. */
IntExecutionContext::IntExecutionContext(string_t name)
    : process_state(psRunnable),
      process_name(name),
      process_priority(50),
      process_error(string_t()),
      notification_message(0),
      process_kind(pkDefault)
{
    if (global_context.get())
        pushNewContext(global_context->clone());
    context_tos = contexts.size();
}

/** Destructor. */
IntExecutionContext::~IntExecutionContext()
{
    // Termination of a process deletes its associated message.
    // Use setNotificationMessage which implements the whole protocol.
    setNotificationMessage(0);

    // Disown all my mutexes
    IntMutex::disownLocksByProcess(this);
}

/** Push new frame. This frame will execute util termination, then execution
    will resume at the current frame. */
void
IntExecutionContext::pushNewFrame(IntExecutionFrame* frame)
{
    frame->context_sp   = contexts.size();
    frame->exception_sp = exceptions.size();
    frame->frame_sp     = frames.size();
    frames.push_back(frame);
    contexts.push_back(new IntExecutionFrameContext(*frame));
}

/** Pop frame. Execution resumes at previous frame. */
void
IntExecutionContext::popFrame()
{
    ASSERT(frames.size() != 0);
    IntExecutionFrame* frame = frames.back();

    /* Clean up stacks */
    while (contexts.size() > frame->context_sp)
        contexts.pop_back();
    while (exceptions.size() > frame->exception_sp)
        exceptions.pop_back();

    /* Generate result */
    if (frame->want_result && frame->bco->isProcedure()) {
        /* Caller wants a result, but we don't have one */
        value_stack.push(0);
    }
    if (!frame->want_result && !frame->bco->isProcedure()) {
        /* Caller doesn't want a result */
        checkStack(1);
        value_stack.pop();
    }

    /* Discard frame */
    frames.pop_back();
}

/** Push exception handler. If program triggers an error, execution resumes
    in the current frame at the specified program counter.
    \param pc Program counter */
void
IntExecutionContext::pushExceptionHandler(IntBytecodeObject::pc_t pc)
{
    IntExceptionHandler* e = new IntExceptionHandler;
    e->value_sp   = value_stack.size();
    e->context_sp = contexts.size();
    e->frame_sp   = frames.size();
    e->pc         = pc;
    exceptions.push_back(e);
}

/** Pop exception handler. Cancels a previous pushExceptionHandler. */
void
IntExecutionContext::popExceptionHandler()
{
    exceptions.pop_back();
}

/** Handle exception. Transfers control to the most recently pushed exception handler,
    if any, and have it process exception e. If there is no active handler, stops
    the bytecode process. */
void
IntExecutionContext::handleException(string_t e, string_t trace)
{
    if (exceptions.size()) {
        /* There is a user-specified exception handler */
        IntExceptionHandler* eh = exceptions.back();

        /* Unwind stacks */
        while (value_stack.size() > eh->value_sp)
            value_stack.pop();
        while (contexts.size() > eh->context_sp)
            contexts.pop_back();
        while (frames.size() > eh->frame_sp)
            frames.pop_back();

        /* Push exception value for user program */
        value_stack.pushNew(new IntStringValue(e));

        /* Change program counter */
        frames.back()->pc = eh->pc;

        /* Pop exception frame */
        exceptions.pop_back();

        /* We may have been called from a suspended process, so make us runnable again */
        if (process_state != psRunning)
            process_state = psRunnable;
    } else {
        /* No user-specified exception handler, so reflect error to caller */
        process_error = IntError(e);
        if (trace.size())
            process_error.addTrace(trace);
        addTraceTo(process_error);
        process_state = psError;
    }
}

/** Copy contexts from a template. */
void
IntExecutionContext::pushContextsFrom(ptr_vector<IntContext>& ctxs)
{
    for (std::size_t i = 0; i < ctxs.size(); ++i) {
        pushNewContext(ctxs.extract_elem(i));
    }
}

string_t
IntExecutionContext::getName() const
{
    return process_name;
}

void
IntExecutionContext::setPriority(int pri)
{
    this->process_priority = pri;
}

int
IntExecutionContext::getPriority() const
{
    return process_priority;
}

const IntError&
IntExecutionContext::getError() const
{
    return process_error;
}


/** Add trace corresponding to last executed instruction to the specified IntError object. */
void
IntExecutionContext::addTraceTo(IntError& err)
{
    static const char*const firstformats[] = {
        N_("in file '%s', line %d"),
        N_("in file '%s'"),
        N_("in %s, file '%s', line %d"),
        N_("in %s")
    };
    static const char*const secondformats[] = {
        N_("called by file '%s', line %d"),
        N_("called by file '%s'"),
        N_("called by %s, file '%s', line %d"),
        N_("called by %s")
    };

    const char*const* formats = firstformats;

    uint32_t context_sp = contexts.size();

    for (uint32_t i = frames.size(); i > 0; --i) {
        /*  file   procedure   line
             -         -         -       -
             -         -         x       -
             -         x         -       "in procedure/function %s"
             -         x         x       "in procedure/function %s"
             x         -         -       "in file %s"
             x         -         x       "in file %s, line %d"
             x         x         -       "in procedure/function %s"
             x         x         x       "in procedure/function %s, file %s, line %d" */
        IntExecutionFrame* frame = frames[i-1];
        IntBCORef bco = frames[i-1]->bco;
        string_t bcoName = bco->getName();
        string_t fileName = Directory::getFileName(bco->getFileName());
        uint32_t lineNr = bco->getLineNumber(frame->pc - 1);

        /* Name the code location */
        if (bcoName.size() == 0) {
            /* No procedure name known. If we know file and or line, list those */
            if (fileName.size()) {
                if (lineNr) {
                    err.addTrace(format(_(formats[0]), fileName, lineNr));
                } else {
                    err.addTrace(format(_(formats[1]), fileName));
                }
                formats = secondformats;
            }
        } else {
            /* Procedure name known. Generate name to use for actual formatting. */
            IntSubroutineValue* sv = dynamic_cast<IntSubroutineValue*>(global_values[global_property_names.lookup(bcoName)]);
            if (sv == 0 || bco != sv->getBytecodeObject()) {
                bcoName = format("(%s)", bcoName);
            }

            if (fileName.size() != 0 && lineNr != 0) {
                err.addTrace(format(_(formats[2]), bcoName, fileName, lineNr));
            } else {
                err.addTrace(format(_(formats[3]), bcoName));
            }
            formats = secondformats;
        }

        /* If we can name the innermost context inside this frame, do so.
           Although we could name all of them, name at most one to keep the
           backtrace concise. */
        while (context_sp > frame->context_sp) {
            --context_sp;
            string_t n = contexts[context_sp]->toString(true);
            if (n.size() && n[0] != '#') {
                err.addTrace(format(_("at %s"), n));
                break;
            }
        }
        context_sp = frame->context_sp;
    }
}

/** Run this process. */
void
IntExecutionContext::run()
{
    unsigned counter = 0;
    process_state = psRunning;
    while (process_state == psRunning) {
        try {
            executeInstruction();
        }
        catch (IntError& e) {
            handleException(e.what(), e.getTrace());
        }
        catch (GError& e) {
            handleException(e.getScriptError(), string_t());
        }
        catch (std::bad_alloc& e) {
            // Do not reflect this back into the script.
            throw;
        }
        catch (std::exception& e) {
            handleException(e.what(), string_t());
        }
        if (++counter == 10000) {
            counter = 0;
            checkForBreak();
        }
    }
}

/** Execute a single instruction. */
void
IntExecutionContext::executeInstruction()
{
    // Terminated?
    if (frames.size() == 0) {
        process_state = psEnded;
        return;
    }

    // Find frame
    IntExecutionFrame& f = *frames.back();

    // Run off end of code?
    if (f.pc >= f.bco->getNumInstructions()) {
        popFrame();
        return;
    }

#if 0
    // Log it
    for (uint32_t i = 0; i < value_stack.size() && i < 10; ++i)
        console << (i ? " : " : "   ") << IntValue::toStringOf(value_stack.top(i), true);
    if (value_stack.size() > 10)
        console << " : ...";
    console << "\n";
    if (process_name.find("Temporary") == string_t::npos) {
        console << process_name << ":" << f.pc << ":   " << f.bco->getDisassembly(f.pc) << "\n";
    }
#endif

    // Execute it
    const IntOpcode& op = (*f.bco)(f.pc++);

    // Cache value_stack address. This saves >3% on my computer.
    IntDataSegment& value_stack = this->value_stack;
    switch (op.major) {
     case IntOpcode::maPush:
        /* Push value */
        switch (op.minor) {
         case IntOpcode::sNamedVariable:
            {
                int32_t index;
                if (IntContext* con = lookup(f.bco->getName(op.arg), index)) {
                    value_stack.pushNew(con->get(index));
                } else {
                    throw IntError::unknownIdentifier(f.bco->getName(op.arg));
                }
            }
            break;
         case IntOpcode::sLocal:
            value_stack.push(f.local_values[op.arg]);
            break;
         case IntOpcode::sStatic:
            value_stack.push(frames.front()->local_values[op.arg]);
            break;
         case IntOpcode::sShared:
            value_stack.push(global_values[op.arg]);
            break;
         case IntOpcode::sNamedShared:
            {
                IntVariableNames::index_t index = global_property_names.lookup(f.bco->getName(op.arg));
                if (index != IntVariableNames::nil)
                    value_stack.push(global_values[index]);
                else
                    throw IntError::unknownIdentifier(f.bco->getName(op.arg));
            }
            break;
         case IntOpcode::sLiteral:
            value_stack.push(f.bco->getLiteral(op.arg));
            break;
         case IntOpcode::sInteger:
            value_stack.pushNew(makeIntValue(int16_t(op.arg)));
            break;
         case IntOpcode::sBoolean:
            value_stack.pushNew(makeBoolValue(int16_t(op.arg)));
            break;
         default:
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maBinary:
        /* Binary operations */
        {
            checkStack(2);
            IntValue* a = value_stack.top(1);
            IntValue* b = value_stack.top(0);
            IntValue* result = executeBinaryOperation(op.minor, a, b);
            value_stack.pop(2);
            value_stack.pushNew(result);
        }
        break;

     case IntOpcode::maUnary:
        /* Unary operations */
        {
            checkStack(1);
            IntValue* result = executeUnaryOperation(op.minor, value_stack.top(0));
            value_stack.pop();
            value_stack.pushNew(result);
        }
        break;

     case IntOpcode::maTernary:
        /* Ternary operations */
        {
            checkStack(3);
            IntValue* a = value_stack.top(2);
            IntValue* b = value_stack.top(1);
            IntValue* c = value_stack.top(0);
            IntValue* result = executeTernaryOperation(op.minor, a, b, c);
            value_stack.pop(3);
            value_stack.pushNew(result);
        }
        break;

     case IntOpcode::maJump:
        /* Jump instructions */
        if ((op.minor & IntOpcode::jOtherMask) != 0) {
            // Other jump
            switch (op.minor & ~IntOpcode::jSymbolic) {
             case IntOpcode::jCatch:
                // "catch"
                pushExceptionHandler(f.bco->getJumpTarget(op.minor, op.arg));
                break;
             case IntOpcode::jDecZero:
                // "jdz", decrement and jump if zero
                // FIXME: optimized execution: when encountering value N here, and
                // there are more than N 'jdz' instructions, jump directly without
                // generating the intermediate results.
                if (handleDecrement())
                    f.pc = f.bco->getJumpTarget(op.minor, op.arg);
                break;
             default:
                handleInvalidOpcode();
                break;
            }
        } else {
            // regular jump
            switch (op.minor & IntOpcode::jAlways) {
             case IntOpcode::jAlways:
                // Jump always
                f.pc = f.bco->getJumpTarget(op.minor, op.arg);
                break;
             case 0:
                // Jump never
                break;
             default:
                // Jump on condition
                checkStack(1);
                int cond = getBoolValue(value_stack.top());
                int mask = (cond < 0 ? IntOpcode::jIfEmpty
                            : cond > 0 ? IntOpcode::jIfTrue
                            : IntOpcode::jIfFalse);
                if ((op.minor & mask) != 0)
                    f.pc = f.bco->getJumpTarget(op.minor, op.arg);
                break;
            }
            if (op.minor & IntOpcode::jPopAlways) {
                checkStack(1);
                value_stack.pop();
            }
        }
        break;

     case IntOpcode::maIndirect:
        /* Indirect call */
        switch (uint8_t operation = op.minor & IntOpcode::miIMOpMask) {
         case IntOpcode::miIMCall:
         case IntOpcode::miIMLoad: {
            checkStack(op.arg + 1);
            std::auto_ptr<IntValue> p(value_stack.extractTop());

            /* CALLIND nargs   rr:args:R      => rr
               PUSHIND nargs   rr:args:R      => rr:result */
            if (p.get() == 0) {
                /* Dereferencing null stays null, but only for functions */
                validateCalledObject(false, op.minor);
                value_stack.pop(op.arg);
                if (operation == IntOpcode::miIMLoad)
                    value_stack.pushNew(0);
            } else if (IntCallableValue* iv = dynamic_cast<IntCallableValue*>(p.get())) {
                /* We can load this */
                validateCalledObject(iv->isProcedureCall(), op.minor);

                IntDataSegment args;
                value_stack.transferTo(op.arg, args);

                iv->call(*this, args, operation == IntOpcode::miIMLoad);
            } else {
                /* Error */
                if (op.minor & IntOpcode::miIMRefuseFunctions)
                    throw IntError::typeError(IntError::tyProcedure);
                else
                    throw IntError::typeError(IntError::tyIndexable);
            }
            break;
         }

         case IntOpcode::miIMStore:
         case IntOpcode::miIMPop:
            /* STOREIND nargs   rr:args:val:R  => rr:val
               POPIND nargs     rr:args:val:R  => rr */
            checkStack(op.arg + 2);
            if (IntIndexableValue* iv = dynamic_cast<IntIndexableValue*>(value_stack.top(0))) {
                /* We can assign this */
                validateCalledObject(iv->isProcedureCall(), op.minor);
                IntValue* value = value_stack.top(1);
                IntArgBlock args(value_stack, value_stack.size() - op.arg - 2, op.arg);
                iv->set(args, value);

                /* Update stack */
                if ((op.minor & IntOpcode::miIMOpMask) == IntOpcode::miIMPop) {
                    value_stack.pop(op.arg+2);
                } else {
                    value_stack.swap(value_stack.size() - op.arg - 2,
                                     value_stack,
                                     value_stack.size() - 2);
                    value_stack.pop(op.arg+1);
                }
            } else {
                /* Error */
                if (op.minor & IntOpcode::miIMRefuseFunctions)
                    throw IntError::typeError(IntError::tyProcedure);
                else
                    throw IntError::typeError(IntError::tyIndexable);
            }
            break;
        }
        break;

     case IntOpcode::maStack:
        /* Stack operations */
        switch (op.minor) {
         case IntOpcode::miStackDup:
            checkStack(op.arg+1);
            value_stack.push(value_stack.top(op.arg));
            break;
         case IntOpcode::miStackDrop:
            checkStack(op.arg);
            value_stack.pop(op.arg);
            break;
         case IntOpcode::miStackSwap:
            checkStack(op.arg+1);
            value_stack.swap(value_stack.size() - op.arg - 1, value_stack, value_stack.size() - 1);
            break;
         default:
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maStore:
        /* Pop or store into variable */
        checkStack(1);
        switch (op.minor) {
         case IntOpcode::sNamedVariable:
            {
                int32_t index;
                if (IntContext* con = lookup(f.bco->getName(op.arg), index)) {
                    con->set(index, value_stack.top());
                } else {
                    throw IntError::unknownIdentifier(f.bco->getName(op.arg));
                }
            }
            break;
         case IntOpcode::sLocal:
            f.local_values.set(op.arg, value_stack.top());
            break;
         case IntOpcode::sStatic:
            frames.front()->local_values.set(op.arg, value_stack.top());
            break;
         case IntOpcode::sShared:
            global_values.set(op.arg, value_stack.top());
            break;
         case IntOpcode::sNamedShared:
            {
                IntVariableNames::index_t index = global_property_names.lookup(f.bco->getName(op.arg));
                if (index != IntVariableNames::nil)
                    global_values.set(index, value_stack.top());
                else
                    throw IntError::unknownIdentifier(f.bco->getName(op.arg));
            }
            break;
         default:
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maPop:
        /* Pop into variable */
        checkStack(1);
        switch (op.minor) {
         case IntOpcode::sNamedVariable:
            {
                int32_t index;
                if (IntContext* con = lookup(f.bco->getName(op.arg), index)) {
                    con->set(index, value_stack.top());
                    value_stack.pop();
                } else {
                    throw IntError::unknownIdentifier(f.bco->getName(op.arg));
                }
            }
            break;
         case IntOpcode::sLocal:
            f.local_values.setNew(op.arg, value_stack.extractTop());
            break;
         case IntOpcode::sStatic:
            frames.front()->local_values.setNew(op.arg, value_stack.extractTop());
            break;
         case IntOpcode::sShared:
            global_values.setNew(op.arg, value_stack.extractTop());
            break;
         case IntOpcode::sNamedShared:
            {
                IntVariableNames::index_t index = global_property_names.lookup(f.bco->getName(op.arg));
                if (index != IntVariableNames::nil)
                    global_values.setNew(index, value_stack.extractTop());
                else
                    throw IntError::unknownIdentifier(f.bco->getName(op.arg));
            }
            break;
         default:
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maMemref:
        switch (op.minor) {
         case IntOpcode::miIMCall:
         case IntOpcode::miIMLoad:
            /* Load/Evaluate TOS.field */
            checkStack(1);
            if (!value_stack.top()) {
                /* Dereferencing null stays null, no change to stack */
            } else if (IntContext* cv = dynamic_cast<IntContext*>(value_stack.top())) {
                /* It's a context */
                int32_t index = cv->lookup(f.bco->getName(op.arg));
                if (index >= 0) {
                    /* Load permitted */
                    IntValue* v = cv->get(index);
                    value_stack.pop();
                    if (op.minor == IntOpcode::miIMCall) {
                        delete v;
                    } else {
                        value_stack.pushNew(v);
                    }
                } else {
                    /* Name not found */
                    throw IntError::unknownIdentifier(f.bco->getName(op.arg));
                }
            } else {
                /* Not a context */
                throw IntError::typeError(IntError::tyRecord);
            }
            break;

         case IntOpcode::miIMStore:
         case IntOpcode::miIMPop:
            /* Store/Pop into TOS.field */
            checkStack(2);
            if (IntContext* cv = dynamic_cast<IntContext*>(value_stack.top())) {
                /* It's a context */
                int32_t index = cv->lookup(f.bco->getName(op.arg));
                if (index >= 0) {
                    /* Assignment permitted */
                    cv->set(index, value_stack.top(1));
                } else {
                    /* Name not found */
                    throw IntError::unknownIdentifier(f.bco->getName(op.arg));
                }
            } else {
                /* Not a context */
                throw IntError::typeError(IntError::tyRecord);
            }
            value_stack.pop();            // context
            if (op.minor == IntOpcode::miIMPop) {
                value_stack.pop();        // value
            }
            break;
         default:
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maDim:
        /* Create variable, initialize with TOS unless it already exists. */
        switch (op.minor) {
         case IntOpcode::sLocal:
            handleDim(f.local_values, f.local_names, op.arg);
            break;
         case IntOpcode::sStatic:
            handleDim(frames.front()->local_values, frames.front()->local_names, op.arg);
            break;
         case IntOpcode::sShared:
            handleDim(global_values, global_property_names, op.arg);
            break;
         default:
            handleInvalidOpcode();
            break;
        }
        break;

     case IntOpcode::maSpecial:
        switch (op.minor) {
         case IntOpcode::miSpecialUncatch:
            /* Cancel previous catch */
            popExceptionHandler();
            break;
         case IntOpcode::miSpecialReturn:
            /* Stop this frame, return to caller */
            checkStack(op.arg);
            popFrame();
            break;
         case IntOpcode::miSpecialWith:
            /* Add TOS to context stack */
            checkStack(1);
            if (IntContext* cv = dynamic_cast<IntContext*>(value_stack.top())) {
                /* Valid context */
                value_stack.extractTop();
                pushNewContext(cv);
            } else {
                /* Not a context */
                throw IntError::typeError(IntError::tyRecord);
            }
            break;
         case IntOpcode::miSpecialEndWith:
            /* Cancel previous miSpecialWith */
            popContext();
            break;
         case IntOpcode::miSpecialFirstIndex:
            /* Start iteration */
            checkStack(1);
            if (IntCallableValue* iv = dynamic_cast<IntCallableValue*>(value_stack.top())) {
                if (IntContext* con = iv->makeFirstContext()) {
                    /* We have something to iterate over */
                    pushNewContext(con);
                    value_stack.pop();
                    value_stack.pushNew(makeBoolValue(1));
                } else {
                    /* Set exists but is empty */
                    value_stack.pop();
                    value_stack.push(0);
                }
            } else {
                /* This is not a set */
                throw IntError::typeError(IntError::tyIterable);
            }
            break;
         case IntOpcode::miSpecialNextIndex:
            /* Continue iteration */
            if (contexts.back()->next()) {
                value_stack.pushNew(makeBoolValue(1));
            } else {
                popContext();
                value_stack.push(0);
            }
            break;
         case IntOpcode::miSpecialEndIndex:
            /* Cancel iteration */
            popContext();
            break;
         case IntOpcode::miSpecialEvalStatement:
            /* Eval */
            handleEvalStatement(op.arg);
            break;
         case IntOpcode::miSpecialEvalExpr:
            /* Eval */
            handleEvalExpression();
            break;
         case IntOpcode::miSpecialDefSub:
            /* Define subroutine */
            {
                checkStack(1);
                IntVariableNames::index_t index = global_property_names.addMaybe(f.bco->getName(op.arg));
                global_values.setNew(index, value_stack.extractTop());
            }
            break;
         case IntOpcode::miSpecialDefShipProperty:
            ship_property_names.addMaybe(f.bco->getName(op.arg));
            break;
         case IntOpcode::miSpecialDefPlanetProperty:
            planet_property_names.addMaybe(f.bco->getName(op.arg));
            break;
         case IntOpcode::miSpecialLoad:
            checkStack(1);
            if (value_stack.top() != 0) {
                string_t name = value_stack.top()->toString(false);
                value_stack.pop();
                if (!handleLoad(name))
                    value_stack.pushNew(new IntStringValue("File not found"));
                else
                    value_stack.push(0);
            }
            break;
         case IntOpcode::miSpecialPrint:
            checkStack(1);
            if (value_stack.top() != 0) {
                console.write(LOG_INFO, value_stack.top()->toString(false));
            }
            value_stack.pop();
            break;
         case IntOpcode::miSpecialAddHook:
            if (op.arg != 0)
                handleInvalidOpcode();
            handleAddHook();
            break;
         case IntOpcode::miSpecialRunHook: {
            if (op.arg != 0)
                handleInvalidOpcode();
            checkStack(1);
            std::auto_ptr<IntValue> p(value_stack.extractTop());
            if (p.get() != 0) {
                string_t name = "ON " + p->toString(false);
                if (IntCallableValue* cv = dynamic_cast<IntCallableValue*>(global_values[global_property_names.lookup(name)])) {
                    /* It is defined and refers to a subroutine (we ignore the invalid case
                       where it is defind but not a subroutine */
                    IntDataSegment args;
                    cv->call(*this, args, false);
                }
            }
            break;
         }

         case IntOpcode::miSpecialThrow:
            checkStack(1);
            if (IntValue* v = value_stack.top())
                handleException(v->toString(false), string_t());
            else
                handleException("Throw empty", string_t());
            break;
         case IntOpcode::miSpecialTerminate:
            process_state = psTerminated;
            break;
         case IntOpcode::miSpecialSuspend:
            process_state = psSuspended;
            break;
         case IntOpcode::miSpecialNewArray:
            handleNewArray(op.arg);
            break;
         case IntOpcode::miSpecialMakeList:
            handleMakeList(op.arg);
            break;
         case IntOpcode::miSpecialNewHash:
            if (op.arg != 0)
                handleInvalidOpcode();
            handleNewHash();
            break;
         case IntOpcode::miSpecialInstance:
            checkStack(1);
            if (IntStructureType* isv = dynamic_cast<IntStructureType*>(value_stack.top())) {
                Ptr<IntStructureTypeData> isvd = isv->getType();
                value_stack.pop();
                value_stack.pushNew(new IntStructureValue(new IntStructureValueData(isvd)));
            } else {
                handleException("Invalid structure constructor", string_t());
            }
            break;
         case IntOpcode::miSpecialResizeArray:
            handleResizeArray(op.arg);
            break;
         case IntOpcode::miSpecialBind:
            handleBind(op.arg);
            break;
         default:
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maFusedUnary:
        if (f.pc < f.bco->getNumInstructions()) {
            IntValue* a = getReferencedValue(op);
            IntValue* result = executeUnaryOperation((*f.bco)(f.pc).minor, a);
            value_stack.pushNew(result);
            ++f.pc;
        } else {
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maFusedBinary:
        if (f.pc < f.bco->getNumInstructions()) {
            checkStack(1);
            IntValue* a = value_stack.top(0);
            IntValue* b = getReferencedValue(op);
            IntValue* result = executeBinaryOperation((*f.bco)(f.pc).minor, a, b);
            value_stack.pop();
            value_stack.pushNew(result);
            ++f.pc;
        } else {
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maFusedComparison:
        /* push + bcmp */
        if (f.pc < f.bco->getNumInstructions()) {
            checkStack(2);
            IntValue* a = value_stack.top(1);
            IntValue* b = value_stack.top(0);

            const IntOpcode& next = (*f.bco)(f.pc);
            int result = executeComparison(op.minor, a, b);
            int mask = (result < 0 ? IntOpcode::jIfEmpty
                        : result == 0 ? IntOpcode::jIfFalse
                        : IntOpcode::jIfTrue);
            if ((next.minor & mask) != 0) {
                // Perform the jump
                f.pc = f.bco->getJumpTarget(next.minor, next.arg);
            } else {
                // Skip the jump
                ++f.pc;
            }
            value_stack.pop(2);
        } else {
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maFusedComparison2:
        /* push + bcmp + jxxp */
        if (f.pc+1 < f.bco->getNumInstructions()) {
            checkStack(1);
            IntValue* a = value_stack.top(0);
            IntValue* b = getReferencedValue(op);

            int result = executeComparison((*f.bco)(f.pc).minor, a, b);
            const IntOpcode& next = (*f.bco)(f.pc+1);
            int mask = (result < 0 ? IntOpcode::jIfEmpty
                        : result == 0 ? IntOpcode::jIfFalse
                        : IntOpcode::jIfTrue);
            if ((next.minor & mask) != 0) {
                // Perform the jump
                f.pc = f.bco->getJumpTarget(next.minor, next.arg);
            } else {
                // Skip the jump
                f.pc += 2;
            }
            value_stack.pop();
        } else {
            handleInvalidOpcode();
        }
        break;

     case IntOpcode::maInplaceUnary:
        /* pushloc + uinc/udec */
        if (f.pc < f.bco->getNumInstructions() && op.minor == IntOpcode::sLocal) {
            /* Check delta */
            int32_t delta = (*f.bco)(f.pc).minor == unInc ? +1 : -1;

            /* Operate */
            IntValue* v = f.local_values.extract(op.arg);
            if (IntIntIntValue* iv = dynamic_cast<IntIntIntValue*>(v)) {
                iv->add(delta);
                ++f.pc;
            } else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(v)) {
                fv->add(delta);
                ++f.pc;
            } else {
                // We cannot operate on this. Just push it.
            }
            value_stack.pushNew(v);
        } else {
            handleInvalidOpcode();
        }
        break;

     default:
        handleInvalidOpcode();
    }
}

/** Suspend this process to perform UI operations. */
void
IntExecutionContext::suspendForUI()
{
    if (process_state == psRunning)
        process_state = psRunnable;
}

/** Set global context for execution. Every newly-created
    IntExecutionContext will have this context active. */
void
IntExecutionContext::setNewGlobalContext(IntContext* p)
{
    global_context.reset(p);
}

/** Set break handler. This function is called periodically to check
    whether the user wants to break. */
void
IntExecutionContext::setBreakHandler(bool check_break())
{
    ::check_break = check_break;
}

/** Perform name lookup.
    \param name  [in] Name to look up
    \param index [out] If found, the context's cookie for access to the name
    \return Context containing the name, 0 if none */
IntContext*
IntExecutionContext::lookup(const string_t& name, int32_t& index)
{
    IntNameQuery q(name);
    for (uint32_t i = contexts.size(); i > 0; --i) {
        IntContext* c = contexts[i-1];
        int n = c->lookup(q);
        if (n >= 0) {
            index = n;
            return c;
        }
    }
    return 0;
}

/** Handle Load instruction. Loads and compiles the file, and pushes an appropriate frame executing it.
    \param name File name given by user
    \retval true file loaded and compiled successfully
    \retval false file not found
    \throw IntError on compilation error */
bool
IntExecutionContext::handleLoad(string_t name)
{
    // Open stream
    Ptr<Stream> ps = game_file_dir->openFileNT(name, Stream::C_READ);
    if (!ps)
        return false;

    // Generate compilation objects
    TextFile tf(*ps);
    IntFileCommandSource fcs(tf);
    IntBCORef nbco = new IntBytecodeObject();
    nbco->setFileName(ps->getName());

    // Compile
    try {
        IntStatementCompiler sc(fcs);
        sc.compileList(*nbco,
                       IntStatementCompilationContext()
                       .withFlags(IntCompilationContext::LocalContext
                                  | IntCompilationContext::ExpressionsAreStatements
                                  | IntCompilationContext::LinearExecution));
        sc.finishBCO(*nbco);
    }
    catch (IntError& e) {
        e.addTrace(format(_("in file '%s', line %d"), Directory::getFileName(ps->getName()), fcs.getLineNumber()));
        throw e;
    }

    // Make new frame
    pushNewFrame(new IntExecutionFrame(nbco));
    return true;
}

/** Set variable in this process. This consumes possible exceptions.
    It is intended to be used by implementations of commands which set
    variables by name; generally, those are "UI.RESULT" or
    "CARGO.REMAINDER" which may not fail.
    \retval true assignment succeeded
    \retval false assignment failed */
bool
IntExecutionContext::setVariable(string_t name, IntValue* value)
{
    try {
        int32_t index;
        if (IntContext* con = lookup(name, index)) {
            con->set(index, value);
            return true;
        }
    }
    catch (IntError&) { }
    return false;
}

/** Get variable in this process.
    \return variable value; Null if variable is not set. */
IntValue*
IntExecutionContext::getVariable(string_t name)
{
    int32_t index;
    IntContext* con = lookup(name, index);
    if (con != 0) {
        return con->get(index);
    } else {
        return 0;
    }
}


/** Get object this process is currently in. */
GObject*
IntExecutionContext::getCurrentObject() const
{
    for (ptr_vector<IntContext>::size_type i = contexts.size(); i > 0; --i)
        if (GObject* obj = contexts[i-1]->getObject())
            return obj;
    return 0;
}

/** Get object this process was invoked from. */
GObject*
IntExecutionContext::getInvokingObject() const
{
    for (ptr_vector<IntContext>::size_type i = context_tos; i > 0; --i)
        if (GObject* obj = contexts[i-1]->getObject())
            return obj;
    return 0;
}

/** Set notification message. The message will be associated with this
    process. In case another message was associated with the process,
    it will be removed.
    \param msg Message. Can be 0 to delete this process' message. */
void
IntExecutionContext::setNotificationMessage(IntNotificationMessage* msg)
{
    if (notification_message != msg) {
        /* Detach old message. Note that oldmsg->remove will call
           setNotificationMessage(0), so first get the message out
           of the way. */
        IntNotificationMessage* oldmsg = notification_message;
        notification_message = 0;
        if (oldmsg != 0) {
            oldmsg->remove();
        }

        /* Attach new message */
        notification_message = msg;
        if (msg != 0) {
            msg->setAssociatedProcess(this);
        }
    }
}

/** Get current notification message. */
IntNotificationMessage*
IntExecutionContext::getNotificationMessage() const
{
    return notification_message;
}

/** Check whether the user wants to break. */
bool
IntExecutionContext::checkForBreak()
{
    /* Check for break, and set termination flag. We cannot just discard
       the executing frames because we're still in executeInstruction which
       still refers to one of them. */
    bool result = check_break();
    if (result)
        process_state = psTerminated;
    return result;
}

/** Check that stack contains a number of values. */
inline void
IntExecutionContext::checkStack(uint32_t required)
{
    if (value_stack.size() < required)
        throw IntError::internalError("stack error");
}

/** Handle invalid opcode. */
void
IntExecutionContext::handleInvalidOpcode()
{
    throw IntError::internalError("invalid opcode");
}

/** Handle Dim instruction.
    \param values data segment affected by this instruction
    \param names name list affected by this instruction
    \param index index into our name table */
void
IntExecutionContext::handleDim(IntDataSegment& values, IntVariableNames& names, uint16_t index)
{
    /* Check stack */
    checkStack(1);

    /* Check name. Must not be blank. */
    const string_t name = frames.back()->bco->getName(index);
    if (name.size() == 0)
        handleInvalidOpcode();

    /* Does it already exist? */
    IntVariableNames::index_t existing = names.lookup(name);
    if (existing == IntVariableNames::nil) {
        /* No, add it */
        values.set(names.add(name), value_stack.top());
    }

    value_stack.pop();
}

/** Handle user function call instruction.
    \param bco called function
    \param args arguments provided by user (can be modified)
    \param want_result true if we expect a result, false if we don't */
void
IntExecutionContext::handleFunctionCall(IntBCORef bco, IntDataSegment& args, bool want_result)
{
    /* Verify number of arguments */
    checkArgumentCount(args.size(), bco->getMinArgs(), bco->isVarargs() ? args.size() : bco->getMaxArgs());

    /* Make a new frame executing this BCO */
    IntExecutionFrame* frame = new IntExecutionFrame(bco);
    pushNewFrame(frame);
    frame->want_result = want_result;

    /* Prepare parameters */
    uint32_t numArgs    = std::min(args.size(), uint32_t(bco->getMaxArgs()));
    uint32_t numVarArgs = args.size() - numArgs;
    Ptr<IntArrayData> va;

    /* Copy the varargs */
    if (bco->isVarargs()) {
        va = new IntArrayData();
        va->addDimension(numVarArgs);
        args.transferTo(numVarArgs, va->content);
    }

    /* Copy regular parameters */
    args.transferTo(numArgs, frame->local_values);

    /* Store the varargs */
    if (bco->isVarargs()) {
        frame->local_values.setNew(bco->getMaxArgs(), new IntArray(va));
    }
}

/** Handle "sevals" statement. Compiles the arguments into a new temporary BCO
    and pushes a frame executing that.
    \param nargs Number of arguments given (=number of script lines) */
void
IntExecutionContext::handleEvalStatement(uint32_t nargs)
{
    /* Verify stack */
    checkStack(nargs);

    /* Build command source */
    IntMemoryCommandSource mcs;
    for (uint32_t i = 0; i < nargs; ++i)
        mcs.addLine(IntValue::toStringOf(value_stack.top(nargs-i-1), false));

    /* Drop args */
    value_stack.pop(nargs);

    /* Prepare compilation */
    IntStatementCompilationContext scc;
    if (nargs == 1) {
        /* One-liner */
        scc.withFlags(scc.LocalContext | scc.ExpressionsAreStatements | scc.RefuseBlocks | scc.LinearExecution);
        scc.withExecutionContext(this);
    } else {
        /* Multi-line block */
        scc.withFlags(scc.LocalContext | scc.ExpressionsAreStatements | scc.LinearExecution);
        scc.withExecutionContext(0);
    }

    /* Compile */
    IntBCORef bco = new IntBytecodeObject();
    IntStatementCompiler sc(mcs);
    sc.compileList(*bco, scc);
    sc.finishBCO(*bco);
    bco->setName("Eval");

    /* Execute */
    pushNewFrame(new IntExecutionFrame(bco));
}

/** Handle "sevalx" instruction. Compiles the stack top into a new BCO,
    and pushes a frame executing that and returning a single result. */
void
IntExecutionContext::handleEvalExpression()
{
    /* Verify stack */
    checkStack(1);

    /* Eval(0) stays 0 */
    if (value_stack.top() == 0)
        return;

    /* Compile */
    IntTokenizer tok(value_stack.top()->toString(false));
    std::auto_ptr<IntExprNode> expr(IntExpressionParser(tok).parse());
    if (tok.getCurrentToken() != tok.tEnd)
        throw IntError::garbageAtEnd(true);
    value_stack.pop();

    IntBCORef bco = new IntBytecodeObject();
    expr->compileValue(*bco, IntCompilationContext());
    bco->setIsProcedure(false);
    bco->addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialReturn, 1);
    optimize(*bco, 1);
    bco->relocate();
    bco->setName("Eval");

    /* Execute */
    IntExecutionFrame* frame = new IntExecutionFrame(bco);
    frame->want_result = true;
    pushNewFrame(frame);
}

/** Handle "saddhook" instruction. */
void
IntExecutionContext::handleAddHook()
{
    /* Verify stack */
    checkStack(2);

    if (value_stack.top() == 0 || value_stack.top(1) == 0) {
        value_stack.pop(2);
        return;
    }

    /* Verify code */
    IntSubroutineValue* sv = dynamic_cast<IntSubroutineValue*>(value_stack.top());
    if (!sv)
        throw IntError::typeError(IntError::tyProcedure);

    /* Verify name */
    string_t hookName = "ON " + value_stack.top(1)->toString(false);
    IntVariableNames::index_t pos = global_property_names.addMaybe(hookName);
    IntBCORef hook;
    if (global_values[pos] == 0) {
        /* Create it */
        hook = new IntBytecodeObject();
        hook->setIsProcedure(true);
        hook->setName(hookName);
        global_values.setNew(pos, new IntSubroutineValue(hook));
    } else {
        /* Use existing */
        IntSubroutineValue* sv = dynamic_cast<IntSubroutineValue*>(global_values[pos]);
        if (!sv)
            throw IntError::typeError(IntError::tyProcedure);
        hook = sv->getBytecodeObject();
    }

    /* Add it */
    hook->addPushLiteral(sv);
    hook->addInstruction(IntOpcode::maIndirect, IntOpcode::miIMCall, 0);

    /* Clean up stack */
    value_stack.pop(2);
}

/** Handle "snewarray" instruction. */
void
IntExecutionContext::handleNewArray(uint16_t ndim)
{
    /* Check preconditions */
    if (ndim == 0)
        handleInvalidOpcode();
    checkStack(ndim);

    /* Create array object */
    Ptr<IntArrayData> ad = new IntArrayData();
    for (uint32_t i = 0; i < ndim; ++i) {
        IntIntValue* iv = dynamic_cast<IntIntValue*>(value_stack.top(ndim - i - 1));
        if (!iv)
            throw IntError::typeError(IntError::tyInteger);
        if (!ad->addDimension(iv->getValue()))
            throw IntError::rangeError();
    }
    value_stack.pop(ndim);
    value_stack.pushNew(new IntArray(ad));
}

/** Handle "sresizearray" instruction. */
void
IntExecutionContext::handleResizeArray(uint16_t ndim)
{
    /* Check preconditions */
    if (ndim == 0)
        handleInvalidOpcode();
    checkStack(ndim+1);

    /* Create dummy array object */
    IntArrayData ad;
    for (uint32_t i = 0; i < ndim; ++i) {
        IntIntValue* iv = dynamic_cast<IntIntValue*>(value_stack.top(ndim - i - 1));
        if (!iv)
            throw IntError::typeError(IntError::tyInteger);
        if (!ad.addDimension(iv->getValue()))
            throw IntError::rangeError();
    }
    value_stack.pop(ndim);

    /* Fetch the array object */
    IntArray* a = dynamic_cast<IntArray*>(value_stack.top());
    if (!a) {
        throw IntError::typeError(IntError::tyArray);
    }
    Ptr<IntArrayData> origAD = a->getData();

    /* Modify it */
    origAD->resize(ad);
    value_stack.pop();
}

/** Handle "smakelist" instruction. */
void
IntExecutionContext::handleMakeList(uint16_t nelems)
{
    /* Check preconditions */
    checkStack(nelems);

    /* Create array object */
    Ptr<IntArrayData> ad = new IntArrayData();
    if (!ad->addDimension(nelems))
        throw IntError::rangeError();

    /* Populate it */
    value_stack.transferTo(nelems, ad->content);
    value_stack.pushNew(new IntArray(ad));
}

/** Handle "snewhash" instruction. */
void
IntExecutionContext::handleNewHash()
{
    Ptr<IntHashData> ad = new IntHashData();
    value_stack.pushNew(new IntHash(ad));
}

/** Handle "sbind" instruction. */
void
IntExecutionContext::handleBind(uint16_t nargs)
{
    /* Check preconditions.
       Note that "sbind 0" is an expensive nop; let's allow it in case it someday allows something clever. */
    checkStack(nargs + 1);

    /* Build the closure */
    std::auto_ptr<IntClosure> c(new IntClosure());
    if (IntCallableValue* func = dynamic_cast<IntCallableValue*>(value_stack.top())) {
        /* OK */
        c->setNewFunction(func);
        value_stack.extractTop();
    } else {
        /* Error */
        throw IntError::typeError(IntError::tyCallable);
    }
    c->addNewArgumentsFrom(value_stack, nargs);

    /* Return */
    value_stack.pushNew(c.release());
}

/** Handle decrement.
    \return true iff result is zero */
bool
IntExecutionContext::handleDecrement()
{
    checkStack(1);
    IntValue* v = value_stack.top();
    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(v)) {
        iv->add(-1);
        return iv->getValue() == 0;
    } else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(v)) {
        fv->add(-1);
        return fv->getValue() == 0;
    } else {
        throw IntError::typeError(IntError::tyNumeric);
    }
}

/** Get value referenced by an instruction. */
IntValue*
IntExecutionContext::getReferencedValue(const IntOpcode& op)
{
    switch (op.minor) {
     case IntOpcode::sLocal:
        return frames.back()->local_values[op.arg];
     case IntOpcode::sStatic:
        return frames.front()->local_values[op.arg];
     case IntOpcode::sShared:
        return global_values[op.arg];
     case IntOpcode::sNamedShared:
        {
            IntVariableNames::index_t index = global_property_names.lookup(frames.back()->bco->getName(op.arg));
            if (index != IntVariableNames::nil) {
                return global_values[index];
            } else {
                throw IntError::unknownIdentifier(frames.back()->bco->getName(op.arg));
            }
        }
     case IntOpcode::sLiteral:
        return frames.back()->bco->getLiteral(op.arg);
     default:
        handleInvalidOpcode();
        return 0;
    }
}
