/**
  *  \file tools/c2pluginw.cc
  *  \brief c2pluginw - Plugin Manager for Win32 Shell Integration
  */

#include <cstdlib>
#include <windows.h>
#include "util/modinit.h"
#include "arch/platform.h"
#include "arch/win32/win32.h"
#include "util/pluginmgr.h"
#include "util/plugin.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "util/except.h"
#include "util/progress.h"
#include "util/pluginstall.h"

/******************************* Utilities *******************************/

const char* progname;

enum MessageType {
    ErrorDialog,
    InfoDialog,
    YesNoDialog
};

static int
say(string_t str, MessageType mt)
{
    // Replace \n -> \r\n for Windows
    string_t::size_type n = 0, p;
    while ((p = str.find('\n', n)) != string_t::npos) {
        str.insert(p, "\r");
        n = p+2;
    }

    // Type?
    UINT type = mt == ErrorDialog
        ? MB_OK | MB_ICONSTOP
        : mt == YesNoDialog
        ? MB_YESNO | MB_ICONQUESTION
        : MB_OK | MB_ICONINFORMATION;

    // Show it
    if (win32::hasUnicodeSupport()) {
        win32::WStr text, title;
        win32::convertToUnicode(text, str, true);
        win32::convertToUnicode(title, _("PCC2 Plugin Installer"), true);
        return MessageBoxW(NULL, &text[0], &title[0], type);
    } else {
        string_t text = win32::convertToANSI(str);
        string_t title = win32::convertToANSI(_("PCC2 Plugin Installer"));
        return MessageBoxA(NULL, text.c_str(), title.c_str(), type);
    }
}

static void
die(string_t str)
{
    say(str, ErrorDialog);
    std::exit(1);
}

static void
createPath(string_t name)
{
    string_t parent = Directory::getDirName(name);
    string_t child  = Directory::getFileName(name);

    Ptr<Directory> d = Directory::openDir(parent);
    Ptr<DirectoryEntry> e = d->getEntryByName(child);
    if (e->getFileType() != DirectoryEntry::FT_DIRECTORY)
        e->createAsDirectory();
}

static bool
checkPreconditions(PluginManager& mgr, Plugin& plug)
{
    // Check for conflicts
    std::vector<Plugin*> plugList;
    mgr.enumConflictingPlugins(plug, plugList);
    if (!plugList.empty()) {
        string_t message = format(_("Plugin '%s' conflicts with the following plugins:"), plug.getId());
        for (size_t i = 0, n = plugList.size(); i < n; ++i) {
            message += "\n";
            message += format("  %s (%s)", plugList[i]->getId(), plugList[i]->getName());
        }
        message += "\n";
        message += _("It cannot be installed.");
        say(message, ErrorDialog);
        return false;
    }

    // Check for preconditions
    Plugin::FeatureSet fset;
    mgr.enumFeatures(fset);
    if (!plug.isSatisfied(fset)) {
        string_t message = format(_("Plugin '%s' requires the following features:"), plug.getId());
        Plugin::FeatureSet missing;
        plug.enumMissingFeatures(fset, missing);
        for (Plugin::FeatureSet::const_iterator it = missing.begin(), e = missing.end(); it != e; ++it) {
            message += "\n  ";
            message += it->first;
            if (!it->second.empty()) {
                message += " ";
                message += it->second;
            }
        }
        message += "\n";
        message += _("It cannot be installed.");
        say(message, ErrorDialog);
        return false;
    }
    return true;
}

static void
doAdd(char** argv)
{
    // Create plugin directory
    string_t pluginDirName = Directory::makeFileName(Platform::getUserSettingsDirectory(1), "plugins");
    createPath(pluginDirName);

    // Create plugin manager
    PluginManager& mgr = PluginManager::getInstance();
    Ptr<Directory> dir = Directory::openDir(pluginDirName);
    mgr.findPlugins(*dir);

    ProgressMonitor pm;
    PluginInstaller installer(mgr, *dir, pm);

    // Iterate
    bool err = false;
    while (char* p = *argv++) {
        try {
            Plugin* plug = installer.prepareInstall(p);
            if (!plug) {
                say(format(_("File '%s' cannot be installed as a plugin. "
                             "A plugin is normally specified with a *.c2p or *.c2z file."), p), ErrorDialog);
                err = true;
            } else {
                bool isUpdate = mgr.getPluginById(plug->getId()) != 0;
                if (checkPreconditions(mgr, *plug)) {
                    const char* tpl = isUpdate
                        ? _("Do you want to update plugin \"%s\" (%s)?)")
                        : _("Do you want to install plugin \"%s\" (%s)?)");
                    string_t message = format(tpl, plug->getName(), plug->getId());
                    if (plug->getDescription().size() > 0) {
                        message += "\n\n";
                        message += plug->getDescription();
                    }
                    if (say(message, YesNoDialog) == IDYES) {
                        installer.doInstall(false);
                        say(format(isUpdate
                                   ? _("Plugin '%s' has been updated.")
                                   : _("Plugin '%s' has been installed."),
                                   plug->getName()),
                            InfoDialog);
                    }
                } else {
                    err = true;
                }
            }
        }
        catch (FileProblemException& e) {
            say(format("%s: %s", e.getFileName(), e.what()), ErrorDialog);
            err = true;
        }
        catch (std::exception& e) {
            say(e.what(), ErrorDialog);
            err = true;
        }
    }
    if (err) {
        std::exit(1);
    }
}

static const ModuleInitializer* c2plugin_modules[] = {
    mod_base_console,
    0
};

int main(int /*argc*/, char** argv)
{
    // Initialize
    performInitialisation(OnStartup, c2plugin_modules);

    argv = Platform::convertArgumentsToUtf8(argv);
    progname = *argv++;
    if (*argv) {
        try {
            doAdd(argv);
        }
        catch (FileProblemException& e) {
            die(e.getFileName() + ": " + e.what());
        }
        catch (std::exception& e) {
            die(e.what());
        }
    } else {
        die(_("Missing plugin file on command line.\n\n"
              "You must invoke this program with one or more *.c2p/*.c2z files on the command line. "
              "You can also use the plugin installer in the PCC2/PlayVCR GUI, or the c2plugin command line tool."));
    }
}
