/**
  *  \file ui/widgets/numbersel.cc
  *  \brief UINumberSelector widget
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/numbersel.h"
#include "ui/window.h"
#include "ui/widgets/statictext.h"
#include "ui/layout.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "client/widgets/helpwidget.h"
#include "ui/widgets/quit.h"

/** Constructor.
    \param id Widget Id
    \param low Lower limit
    \param high Upper limit
    \param set Default step size */
UINumberSelector::UINumberSelector(int id, long low, long high, int step)
    : UIWidget(id), current(low), low(low), high(high), step(step)
{ }

UINumberSelector::~UINumberSelector()
{ }

/** Set value. Checks ranges and clips the value into range. */
void
UINumberSelector::setValue(long val)
{
    if (val < low)
        val = low;
    if (val > high)
        val = high;
    if (val != current) {
        current = val;
        drawWidget(false);
        sig_changed.raise(getId());
    }
}

/** Set range. Clips value into range if it is outside. */
void
UINumberSelector::setRange(int min, int max)
{
    this->low = min;
    this->high = max;
    setValue(current);     /* this will check bounds and update */
}

void
UINumberSelector::increment(int n)
{
    if (n == 0)
        n = 1;
    setSelectedWidget();
    setValue(getValue() + n);
}

void
UINumberSelector::decrement(int n)
{
    if (n == 0)
        n = 1;
    setSelectedWidget();
    setValue(getValue() - n);
}

bool
UINumberSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && hasState(st_Focused)) {
        switch (e.key.code) {
         case '+':
         case SDLK_RIGHT:
            if (int n = e.key.arg)
                increment(n);
            else
                increment(step);
            return true;
         case '-':
         case SDLK_LEFT:
            if (int n = e.key.arg)
                decrement(n);
            else
                decrement(step);
            return true;
         case ss_Ctrl + '+':
         case ss_Ctrl + SDLK_RIGHT:
            increment(100);
            return true;
         case ss_Ctrl + '-':
         case ss_Ctrl + SDLK_LEFT:
            decrement(100);
            return true;
         case ss_Shift + '+':
         case ss_Shift + SDLK_RIGHT:
            increment(1);
            return true;
         case ss_Shift + '-':
         case ss_Shift + SDLK_LEFT:
            decrement(1);
            return true;
         case ss_Alt + '+':
         case ss_Alt + SDLK_RIGHT:
            setSelectedWidget();
            setValue(high);
            return true;
         case ss_Alt + '-':
         case ss_Alt + SDLK_LEFT:
            setSelectedWidget();
            setValue(low);
            return true;
        }
    } else if (e.type == ev_Mouse && e.mouse.b && !hasState(st_Focused) && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        setSelectedWidget();
        setFocusedWidget();
        return true;
    }
    return UIWidget::handleEvent(e, second_pass);
}

/** Standard number selector dialog.
    \param title Dialog title
    \param prompt Prompt displayed before widget
    \return true if change confirmed, false if canceled */
bool
UINumberSelector::doStandardDialog(string_t title, string_t prompt, string_t help)
{
    UIWindow      win(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIStaticText  pr(prompt, UIStaticText::StaticTitle);
    UIButton      plus("+", '+'), minus("-", '-');
    UIButton      ok(1000, _("OK"), SDLK_RETURN);
    UIButton      cancel(1001, _("Cancel"), SDLK_ESCAPE);
    UIButton      helpb("H", 'h');
    UIGroup       g1(UIHBoxLayout::instance5), g2(UIHBoxLayout::instance5);
    UISpacer      spacer;
    UIQuit        quit(1001);
    WHelpWidget   helpw(help);

    ok.addStop();
    cancel.addStop();

    g1.add(minus);
    g1.add(*this);
    g1.add(plus);

    g2.add(ok);
    g2.add(cancel);
    g2.add(spacer);
    if (!help.empty()) {
        g2.add(helpb);
        win.add(helpw);
    }

    win.setLayout(UIVBoxLayout::instance5);
    win.add(pr);
    win.add(g1);
    win.add(g2);
    win.add(quit);
    win.pack();

    setFocusedWidget();

    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    return rv == 1000;
}

